# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import uuid

import pytest

import argilla_sdk as rg
from argilla_sdk.records._export._generic import GenericExportMixin
from argilla_sdk.records._helpers import _dict_to_record


@pytest.fixture
def user_id():
    return str(uuid.uuid4())


@pytest.fixture
def record(user_id):
    return rg.Record(
        fields={"text": "Hello World, how are you?"},
        suggestions=[
            rg.Suggestion("label", "positive", score=0.9),
            rg.Suggestion("topics", ["topic1", "topic2"], score=[0.9, 0.8]),
        ],
        responses=[rg.Response("label", "positive", user_id=user_id)],
        metadata={"source": "twitter", "language": "en"},
        vectors=[rg.Vector("text", [0, 0, 0])],
        id=str(uuid.uuid4()),
    )


def test_export_to_dict(record):
    record_dict = GenericExportMixin._record_to_dict(record)
    imported_record = _dict_to_record(record_dict)

    assert record.responses[0].value == imported_record.responses[0].value
    assert record.suggestions[0].value == imported_record.suggestions[0].value
    for key, value in record.metadata.items():
        assert imported_record.metadata[key] == value
    assert record.fields.text == imported_record.fields.text
    assert record.vectors.text == imported_record.vectors.text


def test_export_to_dict_json(record):
    record_dict = GenericExportMixin._record_to_dict(record)
    record_dict = json.dumps(record_dict)
    record_dict = json.loads(record_dict)
    imported_record = _dict_to_record(record_dict)

    assert record.responses[0].value == imported_record.responses[0].value
    assert record.suggestions[0].value == imported_record.suggestions[0].value
    for key, value in record.metadata.items():
        assert imported_record.metadata[key] == value
    assert record.fields.text == imported_record.fields.text
    assert record.vectors.text == imported_record.vectors.text



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from string import ascii_lowercase

import pytest

import argilla_sdk as rg
from argilla_sdk import Argilla, Dataset, Settings, TextField, Workspace, LabelQuestion


@pytest.fixture
def workspace(client: Argilla) -> Workspace:
    workspace = client.workspaces("test-workspace")
    if not workspace.exists():
        workspace.create()
    yield workspace

    for dataset in workspace.list_datasets():
        client.api.datasets.delete(dataset.id)

    workspace.delete()


@pytest.fixture
def dataset(client: Argilla, workspace: Workspace) -> Dataset:
    name = "".join(random.choices(ascii_lowercase, k=16))
    settings = Settings(
        fields=[TextField(name="text")],
        questions=[LabelQuestion(name="label", labels=["positive", "negative"])],
    )
    dataset = Dataset(
        name=name,
        workspace=workspace.name,
        settings=settings,
        client=client,
    )
    dataset.create()
    yield dataset
    dataset.delete()


def test_query_records_by_text(client: Argilla, dataset: Dataset):
    dataset.records.add(
        [
            {"text": "First record", "id": 1},
            {"text": "Second record", "id": 2},
        ]
    )

    records = list(dataset.records(query="first"))

    assert len(records) == 1
    assert records[0].id == "1"
    assert records[0].fields.text == "First record"

    records = list(dataset.records(query="second"))
    assert len(records) == 1
    assert records[0].id == "2"
    assert records[0].fields.text == "Second record"

    records = list(dataset.records(query="record"))
    assert len(records) == 2


def test_query_records_by_suggestion_value(client: Argilla, dataset: Dataset):
    data = [
        {
            "text": "Hello World, how are you?",
            "label": "positive",
            "id": 1,
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": 2,
        },
        {
            "text": "Hello World, how are you?",
            "label": "positive",
            "id": 3,
        },
    ]

    dataset.records.add(data)

    query = rg.Query(filter=rg.Filter([("label", "==", "positive")]))
    records = list(dataset.records(query=query))

    assert len(records) == 2
    assert records[0].id == "1"
    assert records[1].id == "3"

    query = rg.Query(filter=rg.Filter(("label", "==", "negative")))
    records = list(dataset.records(query=query))

    assert len(records) == 1
    assert records[0].id == "2"

    query = rg.Query(filter=rg.Filter(("label", "in", ["positive", "negative"])))
    records = list(dataset.records(query=query))
    assert len(records) == 3

    test_filter = rg.Filter([("label", "==", "positive"), ("label", "==", "negative")])
    query = rg.Query(filter=test_filter)
    records = list(dataset.records(query=query))
    assert len(records) == 0



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import random
import uuid
from string import ascii_lowercase
from tempfile import TemporaryDirectory

import pytest

import argilla_sdk as rg


@pytest.fixture
def dataset(client) -> rg.Dataset:
    mock_dataset_name = "".join(random.choices(ascii_lowercase, k=16))
    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.TextQuestion(name="label", use_markdown=False),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    dataset.create()
    yield dataset
    dataset.delete()


def test_export_dataset_to_disk(dataset: rg.Dataset):
    mock_data = [
        {
            "text": "Hello World, how are you?",
            "label": "positive",
            "external_id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "external_id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "label": "positive",
            "external_id": uuid.uuid4(),
        },
    ]
    dataset.records.add(records=mock_data)

    with TemporaryDirectory() as temp_dir:
        output_dir = dataset.to_disk(path=temp_dir)

        records_path = os.path.join(output_dir, "records.json")
        assert os.path.exists(records_path)
        with open(records_path, "r") as f:
            exported_records = json.load(f)

        settings_path = os.path.join(output_dir, "settings.json")
        assert os.path.exists(settings_path)
        with open(settings_path, "r") as f:
            exported_settings = json.load(f)

        dataset_path = os.path.join(output_dir, "dataset.json")
        assert os.path.exists(dataset_path)
        with open(dataset_path, "r") as f:
            exported_dataset = json.load(f)

    assert len(exported_records) == len(mock_data)
    assert exported_records[0]["fields"]["text"] == "Hello World, how are you?"
    assert exported_records[0]["suggestions"]["label"]["value"] == "positive"

    assert exported_settings["fields"][0]["name"] == "text"
    assert exported_settings["questions"][0]["name"] == "label"

    assert exported_dataset["name"] == dataset.name


def test_import_dataset_from_disk(dataset: rg.Dataset, client):
    mock_data = [
        {
            "text": "1: Hello World, how are you?",
            "label": "positive",
            "external_id": uuid.uuid4(),
        },
        {
            "text": "2: Hello World, how are you?",
            "label": "negative",
            "external_id": uuid.uuid4(),
        },
        {
            "text": "3: Hello World, how are you?",
            "label": "positive",
            "external_id": uuid.uuid4(),
        },
    ]
    dataset.records.add(records=mock_data)

    with TemporaryDirectory() as temp_dir:
        output_dir = dataset.to_disk(path=temp_dir)
        new_dataset = rg.Dataset.from_disk(output_dir, client=client)

    for i, record in enumerate(new_dataset.records(with_suggestions=True)):
        assert record.fields.text == mock_data[i]["text"]
        assert record.suggestions.label.value == mock_data[i]["label"]

    assert new_dataset.settings.fields[0].name == "text"
    assert new_dataset.settings.questions[0].name == "label"



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

LOG_LEVEL_MAP = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


def log(message: str, level: str = "info") -> None:
    """Log a message at the specified level.
    Args:
        message (str): The message to log.
        level (str): The log level to use. Defaults to "info".
    """
    level_int = LOG_LEVEL_MAP.get(level, logging.INFO)
    logging.log(level=level_int, msg=message)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

from IPython.display import HTML


RESOURCE_REPR_CONFIG = {
    "Dataset": {
        "columns": ["name", "id", "workspace_id", "updated_at"],
        "table_name": "Datasets",
        # "len_column": "records",
    },
    "Workspace": {
        "columns": ["name", "id", "updated_at"],
        "table_name": "Workspaces",
        # "len_column": "datasets",
    },
    "User": {"columns": ["username", "id", "role", "updated_at"], "table_name": "Users"},
}


class ResourceHTMLReprMixin:
    def _resource_to_table_row(self, resource) -> Dict[str, Any]:
        row = {}
        dumped_resource_model = resource._model.model_dump()
        resource_name = resource.__class__.__name__
        config = RESOURCE_REPR_CONFIG[resource_name].copy()
        len_column = config.pop("len_column", None)
        columns = config["columns"]
        if len_column is not None:
            row[len_column] = len(resource)
            columns = [column for column in columns if column != len_column]

        for column in columns:
            row[column] = dumped_resource_model[column]

        return row

    def _resource_to_table_name(self, resource) -> str:
        resource_name = resource.__class__.__name__
        return RESOURCE_REPR_CONFIG[resource_name]["table_name"]

    def _represent_as_html(self, resources) -> HTML:
        table_name = self._resource_to_table_name(resources[0])
        table_rows = [self._resource_to_table_row(resource) for resource in resources]

        html_table = f"<h3>{table_name}</h3><table><tr>"
        for column in table_rows[0]:
            html_table += f"<th>{column}</th>"
        html_table += "</tr>"

        for row in table_rows:
            html_table += "<tr>"
            for column in row:
                html_table += f"<td>{row[column]}</td>"
            html_table += "</tr>"

        html_table += "</table>"
        return HTML(html_table)._repr_html_()



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from datetime import datetime

import httpx
from pytest_httpx import HTTPXMock

import argilla_sdk as rg
from argilla_sdk._models import TextFieldModel


class TestFieldsAPI:
    def test_create_many_fields(self, httpx_mock: HTTPXMock):
        # TODO: Add a test for the delete method in client
        mock_dataset_id = uuid.uuid4()
        mock_return_value = {
            "id": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
            "name": "string",
            "title": "string",
            "required": True,
            "settings": {"type": "text", "use_markdown": False},
            "dataset_id": str(mock_dataset_id),
            "inserted_at": datetime.now().isoformat(),
            "updated_at": datetime.now().isoformat(),
        }
        mock_field = {
            "name": "5044cv0wu5",
            "title": "string",
            "required": True,
            "settings": {"type": "text", "use_markdown": False},
        }
        mock_field = TextFieldModel(**mock_field)
        httpx_mock.add_response(
            json=mock_return_value,
            url=f"http://test_url/api/v1/datasets/{mock_dataset_id}/fields",
            method="POST",
            status_code=200,
        )
        with httpx.Client() as client:
            client = rg.Argilla(api_url="http://test_url")
            client.api.fields.create_many(dataset_id=mock_dataset_id, fields=[mock_field])



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Optional

import httpx

from argilla_sdk._constants import _DEFAULT_API_URL, _DEFAULT_API_KEY


@dataclass
class HTTPClientConfig:
    """Basic configuration for the HTTP client."""

    api_url: Optional[str] = None
    api_key: Optional[str] = None
    timeout: int = None

    def __post_init__(self):
        self.api_url = self.api_url or _DEFAULT_API_URL
        self.api_key = self.api_key or _DEFAULT_API_KEY
        self.timeout = self.timeout or 60


def create_http_client(api_url: str, api_key: str, **client_args) -> httpx.Client:
    """Initialize the SDK with the given API URL and API key."""
    # This piece of code is needed to make old sdk works in combination with new one

    headers = client_args.pop("headers", {})
    headers["X-Argilla-Api-Key"] = api_key

    return httpx.Client(base_url=api_url, headers=headers, **client_args)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union, Iterable
from uuid import UUID, uuid4

from argilla_sdk._models import (
    MetadataModel,
    RecordModel,
    UserResponseModel,
    SuggestionModel,
    VectorModel,
    MetadataValue,
)
from argilla_sdk._resource import Resource
from argilla_sdk.responses import Response, UserResponse
from argilla_sdk.settings import QuestionType, VectorField, TextField, MetadataType
from argilla_sdk.suggestions import Suggestion
from argilla_sdk.vectors import Vector

if TYPE_CHECKING:
    from argilla_sdk.datasets import Dataset


class Record(Resource):
    """The class for interacting with Argilla Records. A `Record` is a single sample
    in a dataset. Records receives feedback in the form of responses and suggestions.
    Records contain fields, metadata, and vectors.

    Attributes:
        id (Union[str, UUID]): The id of the record.
        fields (RecordFields): The fields of the record.
        metadata (RecordMetadata): The metadata of the record.
        vectors (RecordVectors): The vectors of the record.
        responses (RecordResponses): The responses of the record.
        suggestions (RecordSuggestions): The suggestions of the record.
        dataset (Dataset): The dataset to which the record belongs.
        _server_id (UUID): An id for the record generated by the Argilla server.
    """

    _model: RecordModel

    def __init__(
        self,
        id: Optional[Union[UUID, str]] = None,
        fields: Dict[str, Union[str, None]] = None,
        metadata: Optional[Dict[str, MetadataValue]] = None,
        vectors: Optional[List[Vector]] = None,
        responses: Optional[List[Response]] = None,
        suggestions: Optional[List[Suggestion]] = None,
        _server_id: Optional[UUID] = None,
        _dataset: Optional["Dataset"] = None,
    ):
        """Initializes a Record with fields, metadata, vectors, responses, suggestions, external_id, and id.
        Records are typically defined as flat dictionary objects with fields, metadata, vectors, responses, and suggestions
        and passed to Dataset.DatasetRecords.add() as a list of dictionaries.

        Args:
            id: An id for the record. If not provided, a UUID will be generated.
            fields: A dictionary of fields for the record.
            metadata: A dictionary of metadata for the record.
            vectors: A dictionary of vectors for the record.
            responses: A list of Response objects for the record.
            suggestions: A list of Suggestion objects for the record.
            _server_id: An id for the record. (Read-only and set by the server)
            _dataset: The dataset object to which the record belongs.
        """
        self._dataset = _dataset

        self._model = RecordModel(
            fields=fields,
            external_id=id or uuid4(),
            id=_server_id,
        )
        # TODO: All this code blocks could be define as property setters
        # Initialize the fields
        self.__fields = RecordFields(fields=self._model.fields)
        # Initialize the vectors
        self.__vectors = RecordVectors(vectors=vectors, record=self)
        # Initialize the metadata
        self.__metadata = RecordMetadata(metadata=metadata)
        self.__responses = RecordResponses(responses=responses, record=self)
        self.__suggestions = RecordSuggestions(suggestions=suggestions, record=self)

    def __repr__(self) -> str:
        return (
            f"Record(id={self.id},fields={self.fields},metadata={self.metadata},"
            f"suggestions={self.suggestions},responses={self.responses})"
        )

    ############################
    # Properties
    ############################

    @property
    def id(self) -> str:
        return self._model.external_id

    @id.setter
    def id(self, value: str) -> None:
        self._model.external_id = value

    @property
    def dataset(self) -> "Dataset":
        return self._dataset

    @dataset.setter
    def dataset(self, value: "Dataset") -> None:
        self._dataset = value

    @property
    def fields(self) -> "RecordFields":
        return self.__fields

    @property
    def responses(self) -> "RecordResponses":
        return self.__responses

    @property
    def suggestions(self) -> "RecordSuggestions":
        return self.__suggestions

    @property
    def metadata(self) -> "RecordMetadata":
        return self.__metadata

    @property
    def vectors(self) -> "RecordVectors":
        return self.__vectors

    @property
    def _server_id(self) -> Optional[UUID]:
        return self._model.id

    ############################
    # Public methods
    ############################

    def api_model(self) -> RecordModel:
        return RecordModel(
            id=self._model.id,
            external_id=self._model.external_id,
            fields=self.fields.to_dict(),
            metadata=self.metadata.models,
            vectors=self.vectors.models,
            responses=self.responses.api_models(),
            suggestions=self.suggestions.api_models(),
        )

    def serialize(self) -> Dict[str, Any]:
        """Serializes the Record to a dictionary for interaction with the API"""
        serialized_model = self._model.model_dump()
        serialized_suggestions = [suggestion.serialize() for suggestion in self.__suggestions]
        serialized_responses = [response.serialize() for response in self.__responses]
        serialized_model["responses"] = serialized_responses
        serialized_model["suggestions"] = serialized_suggestions
        return serialized_model

    @classmethod
    def from_dict(
        cls,
        dataset: "Dataset",
        data: dict,
        mapping: Optional[Dict[str, str]] = None,
        user_id: Optional[UUID] = None,
    ) -> "Record":
        """Converts a record dictionary to a Record object.
        Args:
            dataset: The dataset object to which the record belongs.
            data: A dictionary representing the record.
            mapping: A dictionary mapping source data keys to Argilla fields, questions, and ids.
            user_id: The user id to associate with the record responses.
        Returns:
            A Record object.
        """
        fields: Dict[str, str] = {}
        responses: List[Response] = []
        record_id: Optional[str] = None
        suggestion_values = defaultdict(dict)
        vectors: List[Vector] = []
        metadata: Dict[str, MetadataValue] = {}

        schema = dataset.schema

        for attribute, value in data.items():
            schema_item = schema.get(attribute)
            attribute_type = None
            sub_attribute = None

            # Map source data keys using the mapping
            if mapping and attribute in mapping:
                attribute_mapping = mapping.get(attribute)
                attribute_mapping = attribute_mapping.split(".")
                attribute = attribute_mapping[0]
                schema_item = schema.get(attribute)
                if len(attribute_mapping) > 1:
                    attribute_type = attribute_mapping[1]
                if len(attribute_mapping) > 2:
                    sub_attribute = attribute_mapping[2]
            elif schema_item is mapping is None and attribute != "id":
                warnings.warn(
                    message=f"""Record attribute {attribute} is not in the schema so skipping. 
                        Define a mapping to map source data fields to Argilla Fields, Questions, and ids
                        """
                )
                continue

            if attribute == "id":
                record_id = value
                continue

            # Add suggestion values to the suggestions
            if attribute_type == "suggestion":
                if sub_attribute in ["score", "agent"]:
                    suggestion_values[attribute][sub_attribute] = value

                elif sub_attribute is None:
                    suggestion_values[attribute].update(
                        {"value": value, "question_name": attribute, "question_id": schema_item.id}
                    )
                else:
                    warnings.warn(
                        message=f"Record attribute {sub_attribute} is not a valid suggestion sub_attribute so skipping."
                    )
                continue

            # Assign the value to question, field, or response based on schema item
            if isinstance(schema_item, TextField):
                fields[attribute] = value
            elif isinstance(schema_item, QuestionType) and attribute_type == "response":
                responses.append(Response(question_name=attribute, value=value, user_id=user_id))
            elif isinstance(schema_item, QuestionType) and attribute_type is None:
                suggestion_values[attribute].update(
                    {"value": value, "question_name": attribute, "question_id": schema_item.id}
                )
            elif isinstance(schema_item, VectorField):
                vectors.append(Vector(name=attribute, values=value))
            elif isinstance(schema_item, MetadataType):
                metadata[attribute] = value
            else:
                warnings.warn(message=f"""Record attribute {attribute} is not in the schema or mapping so skipping.""")
                continue

        suggestions = [Suggestion(**suggestion_dict) for suggestion_dict in suggestion_values.values()]

        return cls(
            id=record_id,
            fields=fields,
            suggestions=suggestions,
            responses=responses,
            vectors=vectors,
            metadata=metadata,
            _dataset=dataset,
        )

    def to_dict(self) -> Dict[str, Dict]:
        """Converts a Record object to a dictionary for export.
        Returns:
            A dictionary representing the record where the keys are "fields",
            "metadata", "suggestions", and "responses". Each field and question is
            represented as a key-value pair in the dictionary of the respective key. i.e.
            `{"fields": {"prompt": "...", "response": "..."}, "responses": {"rating": "..."},
        """
        fields = self.fields.to_dict()
        metadata = dict(self.metadata)
        suggestions = self.suggestions.to_dict()
        responses = self.responses.to_dict()
        vectors = self.vectors.to_dict()
        return {
            "id": self.id,
            "fields": fields,
            "metadata": metadata,
            "suggestions": suggestions,
            "responses": responses,
            "vectors": vectors,
            "_server_id": str(self._model.id),
        }

    @classmethod
    def from_model(cls, model: RecordModel, dataset: "Dataset") -> "Record":
        """Converts a RecordModel object to a Record object.
        Args:
            model: A RecordModel object.
            dataset: The dataset object to which the record belongs.
        Returns:
            A Record object.
        """
        return cls(
            id=model.external_id,
            fields=model.fields,
            metadata={meta.name: meta.value for meta in model.metadata},
            vectors=[Vector.from_model(model=vector) for vector in model.vectors],
            # Responses and their models are not aligned 1-1.
            responses=[
                response
                for response_model in model.responses
                for response in UserResponse.from_model(response_model, dataset=dataset)
            ],
            suggestions=[Suggestion.from_model(model=suggestion, dataset=dataset) for suggestion in model.suggestions],
            _dataset=dataset,
            _server_id=model.id,
        )


class RecordFields:
    """This is a container class for the fields of a Record.
    It allows for accessing fields by attribute and iterating over them.
    """

    def __init__(self, fields: Dict[str, Union[str, None]]) -> None:
        self.__fields = fields or {}
        for key, value in self.__fields.items():
            setattr(self, key, value)

    def __getitem__(self, key: str) -> Optional[str]:
        return self.__fields.get(key)

    def __iter__(self):
        return iter(self.__fields)

    def to_dict(self) -> Dict[str, Union[str, None]]:
        return self.__fields

    def __repr__(self) -> str:
        return self.to_dict().__repr__()


class RecordResponses(Iterable[Response]):
    """This is a container class for the responses of a Record.
    It allows for accessing responses by attribute and iterating over them.
    A record can have multiple responses per question so we set the response
    in a list default dictionary with the question name as the key.
    """

    def __init__(self, responses: List[Response], record: Record) -> None:
        self.record = record
        self.__responses_by_question_name = defaultdict(list)

        self.__responses = responses or []
        for response in self.__responses:
            response.record = self.record
            self.__responses_by_question_name[response.question_name].append(response)

    def api_models(self) -> List[UserResponseModel]:
        """Returns a list of ResponseModel objects."""

        responses_by_user_id = defaultdict(list)
        for response in self.__responses:
            responses_by_user_id[response.user_id].append(response)

        return [
            UserResponse(user_id=user_id, answers=responses, _record=self.record).api_model()
            for user_id, responses in responses_by_user_id.items()
        ]

    def __iter__(self):
        return iter(self.__responses)

    def __getitem__(self, index: int):
        return self.__responses[index]

    def __getattr__(self, name) -> List[Response]:
        return self.__responses_by_question_name[name]

    def to_dict(self) -> Dict[str, List[Dict]]:
        """Converts the responses to a dictionary.
        Returns:
            A dictionary of responses.
        """
        response_dict = defaultdict(list)
        for response in self.__responses:
            response_dict[response.question_name].append({"value": response.value, "user_id": response.user_id})
        return response_dict

    def __repr__(self) -> str:
        return {k: [{"value": v["value"]} for v in values] for k, values in self.to_dict().items()}.__repr__()


class RecordSuggestions(Iterable[Suggestion]):
    """This is a container class for the suggestions of a Record.
    It allows for accessing suggestions by attribute and iterating over them.
    """

    def __init__(self, suggestions: List[Suggestion], record: Record) -> None:
        self.record = record

        self.__suggestions = suggestions or []
        for suggestion in self.__suggestions:
            suggestion.record = self.record
            setattr(self, suggestion.question_name, suggestion)

    def api_models(self) -> List[SuggestionModel]:
        return [suggestion.api_model() for suggestion in self.__suggestions]

    def __iter__(self):
        return iter(self.__suggestions)

    def __getitem__(self, index: int):
        return self.__suggestions[index]

    def to_dict(self) -> Dict[str, List[str]]:
        """Converts the suggestions to a dictionary.
        Returns:
            A dictionary of suggestions.
        """
        suggestion_dict: dict = {}
        for suggestion in self.__suggestions:
            suggestion_dict[suggestion.question_name] = {
                "value": suggestion.value,
                "score": suggestion.score,
                "agent": suggestion.agent,
            }
        return suggestion_dict

    def __repr__(self) -> str:
        return self.to_dict().__repr__()


class RecordVectors:
    """This is a container class for the vectors of a Record.
    It allows for accessing suggestions by attribute and iterating over them.
    """

    def __init__(self, vectors: List[Vector], record: Record) -> None:
        self.__vectors = vectors or []
        self.record = record
        for vector in self.__vectors:
            setattr(self, vector.name, vector.values)

    def __repr__(self) -> str:
        return {vector.name: f"{len(vector.values)}" for vector in self.__vectors}.__repr__()

    @property
    def models(self) -> List[VectorModel]:
        return [vector.api_model() for vector in self.__vectors]

    def to_dict(self) -> Dict[str, List[float]]:
        """Converts the vectors to a dictionary.
        Returns:
            A dictionary of vectors.
        """
        return {vector.name: list(map(float, vector.values)) for vector in self.__vectors}


class RecordMetadata(dict):
    """This is a container class for the metadata of a Record."""

    def __init__(self, metadata: Optional[Dict[str, MetadataValue]] = None) -> None:
        super().__init__(metadata or {})

    def __getattr__(self, item: str):
        return self[item]

    def __setattr__(self, key: str, value: MetadataValue):
        self[key] = value

    @property
    def models(self) -> List[MetadataModel]:
        return [MetadataModel(name=key, value=value) for key, value in self.items()]



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict, Tuple, Union, Optional
from uuid import UUID

import httpx
from typing_extensions import deprecated

from argilla_sdk._api._base import ResourceAPI
from argilla_sdk._exceptions import api_error_handler
from argilla_sdk._models import RecordModel, UserResponseModel, SearchQueryModel

__all__ = ["RecordsAPI"]


class RecordsAPI(ResourceAPI[RecordModel]):
    """Manage datasets via the API"""

    MAX_RECORDS_PER_CREATE_BULK = 500
    MAX_RECORDS_PER_UPSERT_BULK = 500

    http_client: httpx.Client

    ################
    # CRUD methods #
    ################
    @api_error_handler
    def get(self, record_id: UUID) -> RecordModel:
        response = self.http_client.get(f"/api/v1/records/{record_id}")
        response.raise_for_status()
        response_json = response.json()
        return self._model_from_json(response_json=response_json)

    @api_error_handler
    def update(self, record: RecordModel) -> RecordModel:
        response = self.http_client.patch(
            url=f"/api/v1/records/{record.id}",
            json=record.model_dump(),
        )
        response.raise_for_status()
        response_json = response.json()
        return self._model_from_json(response_json=response_json)

    @api_error_handler
    def delete(self, record_id: UUID) -> None:
        response = self.http_client.delete(f"/api/v1/records/{record_id}")
        response.raise_for_status()
        self.log(message=f"Deleted record {record_id}")

    ####################
    # Utility methods #
    ####################
    @api_error_handler
    def list(
        self,
        dataset_id: UUID,
        offset: int = 0,
        limit: int = 100,
        with_suggestions: bool = True,
        with_responses: bool = True,
        with_vectors: Optional[Union[List, bool]] = None,
    ) -> List[RecordModel]:
        """List records in a dataset
        Args:
            dataset_id: The ID of the dataset
            offset: The offset to start from
            limit: The number of records to return
            with_vectors: The name of vectors to include
            with_suggestions: Whether to include suggestions
            with_responses: Whether to include responses
        """
        include = []
        if with_suggestions:
            include.append("suggestions")
        if with_responses:
            include.append("responses")
        if with_vectors:
            include.append(self._represent_vectors_to_include(with_vectors))

        params = {
            "offset": offset,
            "limit": limit,
            "include": include,
        }

        response = self.http_client.get(f"/api/v1/datasets/{dataset_id}/records", params=params)
        response.raise_for_status()
        response_json = response.json()
        json_records = response_json["items"]
        return self._model_from_jsons(json_records)

    @api_error_handler
    def search(
        self,
        dataset_id: UUID,
        query: SearchQueryModel,
        offset: int = 0,
        limit: int = 100,
        with_suggestions: bool = True,
        with_responses: bool = True,
        # TODO: Add support for `with_vectors`
    ) -> Tuple[List[Tuple[RecordModel, float]], int]:
        include = []
        if with_suggestions:
            include.append("suggestions")
        if with_responses:
            include.append("responses")
        params = {
            "offset": offset,
            "limit": limit,
            "include": include,
        }
        response = self.http_client.post(
            f"/api/v1/datasets/{dataset_id}/records/search", json=query.model_dump(by_alias=True), params=params
        )
        response.raise_for_status()
        response_json = response.json()
        json_items = response_json["items"]
        total = response_json["total"]
        return [(self._model_from_json(item["record"]), item["query_score"]) for item in json_items], total

    @api_error_handler
    @deprecated("Use `bulk_create` or `bulk_upsert` instead")
    def create_many(self, dataset_id: UUID, records: List[RecordModel]) -> None:
        record_dicts = [record.model_dump() for record in records]
        response = self.http_client.post(
            url=f"/api/v1/datasets/{dataset_id}/records",
            json={"items": record_dicts},
        )
        response.raise_for_status()
        self.log(message=f"Created {len(records)} records in dataset {dataset_id}")
        # TODO: Once server returns the records, return them here

    @api_error_handler
    @deprecated("Use `bulk_create` or `bulk_upsert` instead")
    def update_many(self, dataset_id: UUID, records: List[RecordModel]) -> None:
        record_dicts = [record.model_dump() for record in records]
        response = self.http_client.patch(
            url=f"/api/v1/datasets/{dataset_id}/records",
            json={"items": record_dicts},
        )
        response.raise_for_status()
        self.log(message=f"Updated {len(records)} records in dataset {dataset_id}")

    @api_error_handler
    def bulk_create(
        self, dataset_id: UUID, records: List[RecordModel]
    ) -> Union[List[RecordModel], Tuple[List[RecordModel], int]]:
        if len(records) > self.MAX_RECORDS_PER_CREATE_BULK:
            raise ValueError(f"Cannot create more than {self.MAX_RECORDS_PER_CREATE_BULK} records at once")
        record_dicts = [record.model_dump() for record in records]
        response = self.http_client.post(
            url=f"/api/v1/datasets/{dataset_id}/records/bulk",
            json={"items": record_dicts},
        )
        response.raise_for_status()
        response_json = response.json()
        self.log(message=f"Created {len(records)} in dataset {dataset_id}")
        return self._model_from_jsons(response_jsons=response_json["items"])

    @api_error_handler
    def bulk_upsert(self, dataset_id: UUID, records: List[RecordModel]) -> Tuple[List[RecordModel], int]:
        if len(records) > self.MAX_RECORDS_PER_UPSERT_BULK:
            raise ValueError(f"Cannot upsert more than {self.MAX_RECORDS_PER_UPSERT_BULK} records at once")
        record_dicts = [record.model_dump() for record in records]
        response = self.http_client.put(
            url=f"/api/v1/datasets/{dataset_id}/records/bulk",
            json={"items": record_dicts},
        )
        response.raise_for_status()
        response_json = response.json()
        updated = len(response_json.get("updated_item_ids", []))
        self.log(
            message=f"Updated {updated} records and create {len(records) - updated} records in dataset {dataset_id}"
        )
        return self._model_from_jsons(response_jsons=response_json["items"]), updated

    ####################
    # Response methods #
    ####################

    @api_error_handler
    def create_record_response(self, record_id: UUID, user_response: UserResponseModel) -> None:
        self.http_client.post(
            url=f"/api/v1/records/{record_id}/responses",
            json=user_response.model_dump(),
        ).raise_for_status()

    def create_record_responses(self, record: RecordModel) -> None:
        if not record.responses:
            return
        if not record.id:
            raise ValueError("Record must have an ID to create responses")
        for record_response in record.responses:
            self.create_record_response(record_id=record.id, user_response=record_response)

    ####################
    # Private methods #
    ####################

    def _model_from_json(self, response_json: Dict) -> RecordModel:
        response_json["inserted_at"] = self._date_from_iso_format(date=response_json["inserted_at"])
        response_json["updated_at"] = self._date_from_iso_format(date=response_json["updated_at"])
        if "vectors" in response_json:
            response_json["vectors"] = [
                {"name": key, "vector_values": value} for key, value in response_json["vectors"].items()
            ]
        return RecordModel(**response_json)

    def _model_from_jsons(self, response_jsons: List[Dict]) -> List[RecordModel]:
        return list(map(self._model_from_json, response_jsons))

    def _represent_vectors_to_include(self, with_vectors: Union[List, str, bool]) -> Union[str, None]:
        """Represent the vectors to include in the API request"""
        vector_stub = "vectors"
        if with_vectors is True:
            return vector_stub
        elif not with_vectors:
            return None
        elif isinstance(with_vectors, str):
            return f"{vector_stub}:{with_vectors}"
        elif isinstance(with_vectors, list):
            return f"{vector_stub}:{','.join(with_vectors)}"
        else:
            raise ValueError(f"Invalid value for with_vectors: {with_vectors}")



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from httpx import Timeout

from argilla_sdk import Argilla


class TestHTTPClient:
    def test_create_default_client(self):
        http_client = Argilla().http_client

        assert http_client is not None
        assert http_client.base_url == "http://localhost:6900"
        assert http_client.timeout == Timeout(60)
        assert http_client.headers["X-Argilla-Api-Key"] == "argilla.apikey"

    def test_create_client_with_custom_timeout(self):
        http_client = Argilla(timeout=30).http_client

        assert http_client is not None
        assert http_client.base_url == "http://localhost:6900"
        assert http_client.timeout == Timeout(30)
        assert http_client.headers["X-Argilla-Api-Key"] == "argilla.apikey"

    def test_create_client_with_custom_api_url(self):
        http_client = Argilla(api_url="http://localhost:8000").http_client

        assert http_client is not None
        assert http_client.base_url == "http://localhost:8000"

    def test_create_client_with_custom_api_key(self):
        http_client = Argilla(api_key="custom.apikey").http_client

        assert http_client is not None
        assert http_client.base_url == "http://localhost:6900"
        assert http_client.headers["X-Argilla-Api-Key"] == "custom.apikey"

    def test_create_client_with_extra_headers(self):
        http_client = Argilla(headers={"X-Custom-Header": "Custom value"}).http_client

        assert http_client is not None
        assert http_client.base_url == "http://localhost:6900"
        assert http_client.headers["X-Argilla-Api-Key"] == "argilla.apikey"
        assert http_client.headers["X-Custom-Header"] == "Custom value"

    def test_create_client_with_extra_cookies(self):
        http_client = Argilla(cookies={"session": "session_id"}).http_client

        assert http_client is not None
        assert http_client.base_url == "http://localhost:6900"
        assert http_client.headers["X-Argilla-Api-Key"] == "argilla.apikey"
        assert http_client.cookies["session"] == "session_id"



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import argilla_sdk as rg


class TestArgilla:
    def test_default_client(self):
        with mock.patch("argilla_sdk.Argilla") as mock_client:
            mock_client.return_value.api_url = "http://localhost:6900"
            mock_client.return_value.api_key = "admin.apikey"
            mock_client.return_value.workspace = "argilla"

            client = rg.Argilla(api_url="http://localhost:6900", api_key="admin.apikey")
            assert client.api_url == "http://localhost:6900"
            assert client.api_key == "admin.apikey"

    def test_multiple_clients(self):
        with mock.patch("argilla_sdk.client._api.APIClient.http_client"):
            local_client = rg.Argilla(api_url="http://localhost:6900", api_key="admin.apikey")
            remote_client = rg.Argilla(api_url="http://argilla.production.net", api_key="admin.apikey")
            assert local_client.api_url == "http://localhost:6900"
            assert remote_client.api_url == "http://argilla.production.net"



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Any, Union, Literal, Annotated

from pydantic import BaseModel, Field


class ResponseFilterScopeModel(BaseModel):
    """Filter scope for filtering on a response entity."""

    entity: Literal["response"] = "response"
    question: Union[str, None] = None
    property: Union[Literal["status"], None] = None


class SuggestionFilterScopeModel(BaseModel):
    """Filter scope for filtering on a suggestion entity."""

    entity: Literal["suggestion"] = "suggestion"
    question: str
    property: Union[Literal["value"], Literal["agent"], Literal["score"], None] = "value"


class MetadataFilterScopeModel(BaseModel):
    """Filter scope for filtering on a metadata entity."""

    entity: Literal["metadata"] = "metadata"
    metadata_property: str


ScopeModel = Annotated[
    Union[
        ResponseFilterScopeModel,
        SuggestionFilterScopeModel,
        MetadataFilterScopeModel,
    ],
    Field(discriminator="entity"),
]


class TermsFilterModel(BaseModel):
    """Filter model for terms filter."""

    type: Literal["terms"] = "terms"
    values: List[str]
    scope: ScopeModel


class RangeFilterModel(BaseModel):
    """Filter model for range filter."""

    type: Literal["range"] = "range"
    ge: Union[Any, None] = None
    le: Union[Any, None] = None
    scope: ScopeModel


FilterModel = Annotated[
    Union[
        TermsFilterModel,
        RangeFilterModel,
    ],
    Field(discriminator="type"),
]


class AndFilterModel(BaseModel):
    """And filter model."""

    type: Literal["and"] = "and"

    and_: List["FilterModel"] = Field(alias="and")


class TextQueryModel(BaseModel):
    """Text query model."""

    q: str
    field: Union[str, None] = None


class QueryModel(BaseModel):
    """Query part of the search query model"""

    text: TextQueryModel


class SearchQueryModel(BaseModel):
    """The main search query model."""

    query: Union[QueryModel, None] = None
    filters: Union[AndFilterModel, None] = None



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import uuid

import pytest

from argilla_sdk import Argilla, Dataset, Settings, TextField, RatingQuestion


@pytest.fixture(scope="session", autouse=True)
def clean_datasets(client: Argilla):
    datasets = client.datasets
    for dataset in datasets:
        if dataset.name.startswith("test_"):
            dataset.delete()
    yield


class TestCreateDatasets:
    def test_create_dataset(self, client: Argilla):
        dataset_name = f"test_dataset_{uuid.uuid4()}"
        dataset = Dataset(
            name=dataset_name,
            settings=Settings(
                fields=[TextField(name="test_field")],
                questions=[RatingQuestion(name="test_question", values=[1, 2, 3, 4, 5])],
            ),
        )
        client.datasets.add(dataset)

        assert dataset in client.datasets
        assert dataset.exists()

        created_dataset = client.datasets(name=dataset_name)
        assert created_dataset.settings == dataset.settings

    def test_create_multiple_dataset_with_same_settings(self, client: Argilla):
        dataset_name = f"test_dataset_{uuid.uuid4()}"

        settings = Settings(
            fields=[TextField(name="text")],
            questions=[RatingQuestion(name="question", values=[1, 2, 3, 4, 5])],
        )
        dataset = Dataset(name=dataset_name, settings=settings, client=client).create()
        dataset2 = Dataset(name=f"{dataset_name}_2", settings=settings, client=client).create()

        assert dataset in client.datasets
        assert dataset2 in client.datasets

        assert dataset.exists()
        assert dataset2.exists()

        for ds in [dataset, dataset2]:
            schema = client.datasets(name=ds.name).schema

            assert isinstance(schema["text"], TextField)
            assert schema["text"].name == "text"
            assert isinstance(schema["question"], RatingQuestion)
            assert schema["question"].name == "question"
            assert schema["question"].values == [1, 2, 3, 4, 5]

    def test_create_dataset_from_existing_dataset(self, client: Argilla):
        dataset_name = f"test_dataset_{uuid.uuid4()}"
        dataset = Dataset(
            name=dataset_name,
            settings=Settings(
                fields=[TextField(name="text")],
                questions=[RatingQuestion(name="question", values=[1, 2, 3, 4, 5])],
            ),
        ).create()

        assert dataset in client.datasets
        created_dataset = client.datasets(dataset.name)

        dataset_copy = Dataset(name=f"{dataset.name}_copy", settings=created_dataset.settings, client=client).create()
        assert dataset_copy in client.datasets

        schema = dataset_copy.schema
        assert isinstance(schema["text"], TextField)
        assert schema["text"].name == "text"
        assert isinstance(schema["question"], RatingQuestion)
        assert schema["question"].name == "question"
        assert schema["question"].values == [1, 2, 3, 4, 5]



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argilla_sdk import Argilla, Workspace


class TestWorkspacesManagement:
    def test_create_workspace(self, client: Argilla):
        workspace = Workspace(name="test_workspace")
        client.workspaces.add(workspace)

        assert workspace in client.workspaces
        assert workspace.exists()

    def test_create_and_delete_workspace(self, client: Argilla):
        workspace = client.workspaces(name="test_workspace")
        if workspace.exists():
            for dataset in workspace.datasets:
                dataset.delete()
            workspace.delete()

        workspace.create()
        assert workspace.exists()

        workspace.delete()
        assert not workspace.exists()

    def test_add_and_remove_users_to_workspace(self, client: Argilla):
        workspace = client.workspaces(name="test_workspace")

        test_user = client.users(username="test_user")
        if test_user.exists():
            test_user.delete()

        workspace.create()

        test_user.password = "test_password"
        test_user.create()

        user = workspace.add_user(user=test_user.username)
        assert user in workspace.users

        user = workspace.remove_user(user=user)
        assert user not in workspace.users



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Optional

import httpx

from argilla_sdk._api import HTTPClientConfig, create_http_client
from argilla_sdk._api._datasets import DatasetsAPI
from argilla_sdk._api._fields import FieldsAPI
from argilla_sdk._api._metadata import MetadataAPI
from argilla_sdk._api._questions import QuestionsAPI
from argilla_sdk._api._records import RecordsAPI
from argilla_sdk._api._users import UsersAPI
from argilla_sdk._api._vectors import VectorsAPI
from argilla_sdk._api._workspaces import WorkspacesAPI
from argilla_sdk._constants import _DEFAULT_API_KEY, _DEFAULT_API_URL

__all__ = ["APIClient"]


ARGILLA_API_URL = os.getenv(key="ARGILLA_API_URL", default=_DEFAULT_API_URL)
ARGILLA_API_KEY = os.getenv(key="ARGILLA_API_KEY", default=_DEFAULT_API_KEY)
DEFAULT_HTTP_CONFIG = HTTPClientConfig(api_url=ARGILLA_API_URL, api_key=ARGILLA_API_KEY)


class ArgillaAPI:
    """Argilla API access object."""

    def __init__(self, http_client: httpx.Client):
        self.http_client = http_client

        self.__workspaces = WorkspacesAPI(http_client=self.http_client)
        self.__datasets = DatasetsAPI(http_client=self.http_client)
        self.__users = UsersAPI(http_client=self.http_client)
        self.__fields = FieldsAPI(http_client=self.http_client)
        self.__questions = QuestionsAPI(http_client=self.http_client)
        self.__records = RecordsAPI(http_client=self.http_client)
        self.__vectors = VectorsAPI(http_client=self.http_client)
        self.__metadata = MetadataAPI(http_client=self.http_client)

    @property
    def workspaces(self) -> "WorkspacesAPI":
        return self.__workspaces

    @property
    def users(self) -> "UsersAPI":
        return self.__users

    @property
    def datasets(self) -> "DatasetsAPI":
        return self.__datasets

    @property
    def fields(self) -> "FieldsAPI":
        return self.__fields

    @property
    def questions(self) -> "QuestionsAPI":
        return self.__questions

    @property
    def records(self) -> "RecordsAPI":
        return self.__records

    @property
    def vectors(self) -> "VectorsAPI":
        return self.__vectors

    @property
    def metadata(self) -> "MetadataAPI":
        return self.__metadata


class APIClient:
    """Initialize the SDK with the given API URL and API key.
    This class is used to create an instance of the Argilla API client.

    Args:
        api_url (str, optional): The URL of the Argilla API. Defaults to the value of
            the `ARGILLA_API_URL` environment variable.
        api_key (str, optional): The API key to authenticate with the Argilla API. Defaults to
            the value of the `ARGILLA_API_KEY` environment variable.
        timeout (int, optional): The timeout in seconds for the HTTP requests. Defaults to 60.
        **http_client_args: Additional keyword arguments to pass to the httpx.Client instance.
            See https://www.python-httpx.org/api/#client for more information.
    """

    def __init__(
        self,
        api_url: Optional[str] = DEFAULT_HTTP_CONFIG.api_url,
        api_key: Optional[str] = DEFAULT_HTTP_CONFIG.api_key,
        timeout: int = DEFAULT_HTTP_CONFIG.timeout,
        **http_client_args,
    ):
        http_client_args = http_client_args or {}
        http_client_args["timeout"] = timeout

        self.api_url = api_url
        self.api_key = api_key
        self._http_client_args = http_client_args

    @property
    def http_client(self) -> httpx.Client:
        return create_http_client(
            api_url=self.api_url,  # type: ignore
            api_key=self.api_key,  # type: ignore
            **self._http_client_args,
        )

    @property
    def api(self) -> "ArgillaAPI":
        return ArgillaAPI(http_client=self.http_client)

    ##############################
    # Utility methods
    ##############################

    def log(self, message: str, level: int = logging.INFO) -> None:
        class_name = self.__class__.__name__
        message = f"{class_name}: {message}"
        logging.log(level=level, msg=message)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import fields, dataclass
from typing import Type, TypeVar

__all__ = ["dataclass_instance_from_dict"]

T = TypeVar("T", bound=dataclass)


def dataclass_instance_from_dict(cls: Type[T], data: dict) -> T:
    """Create a dataclass instance from a dictionary, ignoring extra keys found in the dictionary."""

    field_names = set(f.name for f in fields(cls))
    return cls(**{k: v for k, v in data.items() if k in field_names})



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import List, Literal, Optional, Union, Annotated
from uuid import UUID

from pydantic import BaseModel, Field, field_serializer, field_validator, model_validator

from argilla_sdk._exceptions import MetadataError


class MetadataPropertyType(str, Enum):
    terms = "terms"
    integer = "integer"
    float = "float"


class BaseMetadataPropertySettings(BaseModel):
    type: MetadataPropertyType
    visible_for_annotators: Optional[bool] = True


class TermsMetadataPropertySettings(BaseMetadataPropertySettings):
    type: Literal[MetadataPropertyType.terms]
    values: Optional[List[str]] = None

    @field_validator(
        "values",
    )
    @classmethod
    def __validate_values(cls, values):
        if values is None:
            return None
        if not isinstance(values, list):
            raise ValueError(f"values must be a list, got {type(values)}")
        elif not all(isinstance(value, str) for value in values):
            raise ValueError("All values must be strings for terms metadata.")
        return values


class NumericMetadataPropertySettings(BaseMetadataPropertySettings):
    min: Optional[Union[int, float]] = None
    max: Optional[Union[int, float]] = None

    @model_validator(mode="before")
    @classmethod
    def __validate_min_max(cls, values):
        min_value = values.get("min")
        max_value = values.get("max")

        if min_value is not None and max_value is not None:
            if min_value >= max_value:
                raise MetadataError("min must be less than max.")
        return values


class IntegerMetadataPropertySettings(NumericMetadataPropertySettings):
    type: Literal[MetadataPropertyType.integer]

    @model_validator(mode="before")
    @classmethod
    def __validate_min_max(cls, values):
        min_value = values.get("min")
        max_value = values.get("max")

        if not all(isinstance(value, int) or value is None for value in [min_value, max_value]):
            raise MetadataError("min and max must be integers.")
        return values


class FloatMetadataPropertySettings(NumericMetadataPropertySettings):
    type: Literal[MetadataPropertyType.float]


MetadataPropertySettings = Annotated[
    Union[
        TermsMetadataPropertySettings,
        IntegerMetadataPropertySettings,
        FloatMetadataPropertySettings,
    ],
    Field(..., discriminator="type"),
]


class MetadataFieldModel(BaseModel):
    """The schema definition of a metadata field in an Argilla dataset."""

    id: Optional[UUID] = None
    name: str
    settings: MetadataPropertySettings

    type: Optional[MetadataPropertyType] = Field(None, validate_default=True)
    title: Optional[str] = None
    visible_for_annotators: Optional[bool] = True

    @field_validator("name")
    @classmethod
    def __name_lower(cls, name):
        formatted_name = name.lower().replace(" ", "_")
        return formatted_name

    @field_validator("title")
    @classmethod
    def __title_default(cls, title, values):
        validated_title = title or values.data["name"]
        return validated_title

    @field_serializer("id", when_used="unless-none")
    def serialize_id(self, value: UUID) -> str:
        return str(value)

    @field_validator("type", mode="plain")
    @classmethod
    def __validate_type(cls, type, values):
        if type is None:
            return values.data["settings"].type
        return type



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, TYPE_CHECKING, List, Dict, Optional, Iterable
from uuid import UUID

from argilla_sdk._models import UserResponseModel, ResponseStatus
from argilla_sdk._resource import Resource
from argilla_sdk.settings import RankingQuestion

if TYPE_CHECKING:
    from argilla_sdk import Argilla, Dataset, Record

__all__ = ["Response", "UserResponse"]


class Response:
    """Class for interacting with Argilla Responses of records. Responses are answers to questions by a user.
    Therefore, a recod question can have multiple responses, one for each user that has answered the question.
    A `Response` is typically created by a user in the UI or consumed from a data source as a label,
    unlike a `Suggestion` which is typically created by a model prediction.

    """

    def __init__(
        self,
        question_name: str,
        value: Any,
        user_id: UUID,
        _record: Optional["Record"] = None,
    ) -> None:
        """Initializes a `Response` for a `Record` with a user_id and value"""

        if question_name is None:
            raise ValueError("question_name is required")
        if value is None:
            raise ValueError("value is required")
        if user_id is None:
            raise ValueError("user_id is required")

        self.record = _record
        self.question_name = question_name
        self.value = value
        self.user_id = user_id

    def serialize(self) -> dict[str, Any]:
        """Serializes the Response to a dictionary. This is principally used for sending the response to the API, \
            but can be used for data wrangling or manual export.
        
        Returns:
            dict[str, Any]: The serialized response as a dictionary with keys `question_name`, `value`, and `user_id`.
            
        Examples:
        
        ```python
        response = rg.Response("label", "negative", user_id=user.id)
        response.serialize()
        ```
        """
        return {
            "question_name": self.question_name,
            "value": self.value,
            "user_id": self.user_id,
        }

    #####################
    # Private Interface #
    #####################


class UserResponse(Resource):
    """
    Class for interacting with Argilla User Responses of records.  The UserResponse class is a collection
    of responses to questions for a given user. UserResponses are typically created by a user in the UI and
    are defined by ingesting a list of responses from a third-party data source.

    In most cases users will interact with the `UserResponse` class through the `Record` class when
    collected from the server or when creating new records.

    Attributes:
        status (ResponseStatus): The status of the UserResponse (draft, submitted, etc.)
        user_id (UUID): The user_id of the UserResponse (the user who answered the questions)
        answers (List[Response]): A list of responses to questions for the user

    """

    _model: UserResponseModel

    def __init__(
        self,
        user_id: UUID,
        answers: List[Response],
        status: ResponseStatus = "draft",
        client: Optional["Argilla"] = None,
        _record: Optional["Record"] = None,
    ) -> None:
        """Initializes a UserResponse with a user and a set of question answers"""

        super().__init__(client=client)

        self._record = _record
        self._model = UserResponseModel(
            values=self.__responses_as_model_values(answers),
            status=status,
            user_id=user_id,
        )

    def __iter__(self) -> Iterable[Response]:
        return iter(self.answers)

    @property
    def status(self) -> ResponseStatus:
        """Returns the status of the UserResponse"""
        return self._model.status

    @status.setter
    def status(self, status: ResponseStatus) -> None:
        """Sets the status of the UserResponse"""
        self._model.status = status

    @property
    def user_id(self) -> UUID:
        """Returns the user_id of the UserResponse"""
        return self._model.user_id

    @user_id.setter
    def user_id(self, user_id: UUID) -> None:
        """Sets the user_id of the UserResponse"""
        self._model.user_id = user_id

    @property
    def answers(self) -> List[Response]:
        """Returns the list of responses"""
        return self.__model_as_response_list(self._model)

    @classmethod
    def from_model(cls, model: UserResponseModel, dataset: "Dataset") -> "UserResponse":
        """Creates a UserResponse from a ResponseModel"""
        answers = cls.__model_as_response_list(model)

        for answer in answers:
            question = dataset.settings.question_by_name(answer.question_name)
            if isinstance(question, RankingQuestion):
                answer.value = cls.__ranking_from_model_value(answer.value)  # type: ignore

        return cls(user_id=model.user_id, answers=answers, status=model.status)

    def api_model(self):
        """Returns the model that is used to interact with the API"""

        values = self.__responses_as_model_values(self.answers)
        for question_name, value in values.items():
            question = self._record.dataset.settings.question_by_name(question_name)
            if isinstance(question, RankingQuestion):
                value["value"] = self.__ranking_to_model_value(value["value"])

        return UserResponseModel(values=values, status=self._model.status, user_id=self._model.user_id)

    def to_dict(self) -> Dict[str, Any]:
        """Returns the UserResponse as a dictionary"""
        return self._model.dict()

    @staticmethod
    def __responses_as_model_values(answers: List[Response]) -> Dict[str, Dict[str, Any]]:
        """Creates a dictionary of response values from a list of Responses"""
        return {answer.question_name: {"value": answer.value} for answer in answers}

    @staticmethod
    def __model_as_response_list(model: UserResponseModel) -> List[Response]:
        """Creates a list of Responses from a UserResponseModel"""
        return [
            Response(question_name=question_name, value=value["value"], user_id=model.user_id)
            for question_name, value in model.values.items()
        ]

    @classmethod
    def __ranking_from_model_value(cls, value: List[Dict[str, Any]]) -> List[str]:
        return [v["value"] for v in value]

    @classmethod
    def __ranking_to_model_value(cls, value: List[str]) -> List[Dict[str, str]]:
        return [{"value": v} for v in value]



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argilla_sdk import Argilla, Dataset, Settings, TextField, TextQuestion, Workspace


class TestDatasetsList:
    def test_list_datasets(self, client: Argilla):
        workspace = Workspace(name="test_workspace", client=client)
        workspace.create()

        dataset = Dataset(
            name="test_dataset",
            workspace=workspace.name,
            settings=Settings(fields=[TextField(name="text")], questions=[TextQuestion(name="text_question")]),
            client=client,
        )
        dataset.create()
        datasets = client.datasets
        assert len(datasets) > 0, "No datasets were found"

        for ds in datasets:
            if ds.name == "test_dataset":
                assert ds == dataset, "The dataset was not loaded properly"



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, List, Any, Union, Tuple

from argilla_sdk._models import SearchQueryModel
from argilla_sdk._models._search import (
    TextQueryModel,
    ResponseFilterScopeModel,
    SuggestionFilterScopeModel,
    MetadataFilterScopeModel,
    ScopeModel,
    RangeFilterModel,
    TermsFilterModel,
    FilterModel,
    AndFilterModel,
    QueryModel,
)


class Condition(Tuple[str, str, Any]):
    """This class is used to map user conditions to the internal filter models"""

    @property
    def model(self) -> FilterModel:
        field, operator, value = self

        field = field.strip()
        scope = self._extract_filter_scope(field)

        operator = operator.strip()
        if operator == "==":
            return TermsFilterModel(values=[value], scope=scope)
        elif operator == "in":
            return TermsFilterModel(values=value, scope=scope)
        elif operator in [">="]:
            return RangeFilterModel(ge=value, scope=scope)
        elif operator == "<=":
            return RangeFilterModel(le=value, scope=scope)
        else:
            raise ValueError(f"Unknown operator: {operator}")

    @staticmethod
    def _extract_filter_scope(field: str) -> ScopeModel:
        field = field.strip()

        if field == "status":
            return ResponseFilterScopeModel(property="status")
        elif "metadata" in field:
            _, md_property = field.split(".")
            return MetadataFilterScopeModel(metadata_property=md_property)
        elif "suggestion" in field:
            question, _ = field.split(".")
            return SuggestionFilterScopeModel(question=question)
        elif "score" in field:
            question, _ = field.split(".")
            return SuggestionFilterScopeModel(question=question, property="score")
        elif "response" in field:
            question, _ = field.split(".")
            return ResponseFilterScopeModel(question=question)
        else:  # Question field -> Suggestion
            return SuggestionFilterScopeModel(question=field)


class Filter:
    """This class is used to map user filters to the internal filter models"""

    def __init__(self, conditions: Union[List[Tuple[str, str, Any]], Tuple[str, str, Any], None] = None):
        """ Create a filter object for use in Argilla search requests.
        
        Parameters:
            conditions (Union[List[Tuple[str, str, Any]], Tuple[str, str, Any], None], optional): \
                The conditions that will be used to filter the search results. \
                The conditions should be a list of tuples where each tuple contains \
                the field, operator, and value. For example `("label", "in", ["positive","happy"])`.\
        
        """

        if isinstance(conditions, tuple):
            conditions = [conditions]
        self.conditions = [Condition(condition) for condition in conditions]

    @property
    def model(self) -> AndFilterModel:
        return AndFilterModel.parse_obj({"and": [condition.model for condition in self.conditions]})


class Query:
    """This class is used to map user queries to the internal query models"""

    query: Optional[str] = None

    def __init__(self, *, query: Union[str, None] = None, filter: Union[Filter, None] = None):
        """Create a query object for use in Argilla search requests.add()

        Parameters:
            query (Union[str, None], optional): The query string that will be used to search.
            filter (Union[Filter, None], optional): The filter object that will be used to filter the search results.
        """

        self.query = query
        self.filter = filter

    @property
    def model(self) -> SearchQueryModel:
        model = SearchQueryModel()

        if self.query is not None:
            text_query = TextQueryModel(q=self.query)
            model.query = QueryModel(text=text_query)

        if self.filter is not None:
            model.filters = self.filter.model

        return model


__all__ = ["Query", "Filter", "Condition"]



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Any, TYPE_CHECKING, Optional
from uuid import UUID

from argilla_sdk._exceptions import ArgillaSerializeError
from argilla_sdk._helpers._mixins import LoggingMixin, UUIDMixin

if TYPE_CHECKING:
    from argilla_sdk.client import Argilla
    from argilla_sdk._models import ResourceModel
    from argilla_sdk._api._base import ResourceAPI


class Resource(LoggingMixin, UUIDMixin):
    """Base class for all resources (Dataset, Workspace, User, etc.)"""

    _model: "ResourceModel"
    _client: "Argilla"
    _api: "ResourceAPI"

    _MAX_OUTDATED_RETENTION = 30

    def __init__(self, api: Optional["ResourceAPI"] = None, client: Optional["Argilla"] = None) -> None:
        self._client = client
        self._api = api

        self._last_api_call = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._model})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, Resource):
            return False
        if not hasattr(other, "_model"):
            return super().__eq__(other)
        return self._model == other._model

    @property
    def id(self) -> Optional[UUID]:
        return self._model.id

    @id.setter
    def id(self, value: UUID) -> None:
        self._model.id = value

    @property
    def is_outdated(self) -> bool:
        """Checks if the resource is outdated based on the last API call
        Returns:
            bool: True if the resource is outdated, False otherwise
        """
        seconds = self._seconds_from_last_api_call()
        if seconds is None:
            return True
        return seconds > self._MAX_OUTDATED_RETENTION

    def api_model(self):
        """Returns the model that is used to interact with the API"""
        return self._model

    ############################
    # CRUD operations
    ############################

    def create(self) -> "Resource":
        response_model = self._api.create(self._model)
        self._sync(response_model)
        self._update_last_api_call()
        self.log(f"Resource created: {self}")
        return self

    def get(self) -> "Resource":
        response_model = self._api.get(self._model.id)
        self._sync(response_model)
        self._update_last_api_call()
        self.log(f"Resource fetched: {self}")
        return self

    def update(self) -> "Resource":
        response_model = self._api.update(self._model)
        self._sync(response_model)
        self._update_last_api_call()
        self.log(f"Resource updated: {self}")
        return self

    def delete(self) -> None:
        self._api.delete(self._model.id)
        self._update_last_api_call()
        self.log(f"Resource deleted: {self}")

    ############################
    # Serialization
    ############################

    def serialize(self) -> dict[str, Any]:
        try:
            return self._model.model_dump()
        except Exception as e:
            raise ArgillaSerializeError(f"Failed to serialize the resource. {e.__class__.__name__}") from e

    def serialize_json(self) -> str:
        try:
            return self._model.model_dump_json()
        except Exception as e:
            raise ArgillaSerializeError(f"Failed to serialize the resource. {e.__class__.__name__}") from e

    def _sync(self, model: "ResourceModel"):
        """Updates the resource with the ClientAPI that is used to interact with
        Argilla and adds an updated model to the resource.
        Args:
            model (Union[WorkspaceModel, UserModel, DatasetModel]): The updated model
        Returns:
            Self: The updated resource
        """
        self._model = model
        # set all attributes from the model to the resource
        for field in self._model.model_fields:
            setattr(self, field, getattr(self._model, field))
        return self

    def _update_last_api_call(self):
        self._last_api_call = datetime.utcnow()

    def _seconds_from_last_api_call(self) -> Optional[float]:
        if self._last_api_call:
            return (datetime.utcnow() - self._last_api_call).total_seconds()



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from datetime import datetime

import httpx
import pytest
from pytest_httpx import HTTPXMock

import argilla_sdk as rg
from argilla_sdk._exceptions import (
    BadRequestError,
    ConflictError,
    ForbiddenError,
    InternalServerError,
    NotFoundError,
    UnprocessableEntityError,
)


class TestWorkspacesSerialization:
    def test_serialize(self):
        ws = rg.Workspace(
            name="test-workspace",
            id=uuid.uuid4(),
        )

        assert ws.name == ws.serialize()["name"]

    def test_json_serialize_raise_typeerror(self):
        with pytest.raises(TypeError):
            rg.Workspace(
                name="test-workspace",
                id=uuid.uuid4(),
                inserted_at=datetime.utcnow(),
                updated_at=datetime.utcnow(),
            )


class TestWorkspaces:
    @pytest.mark.parametrize(
        "status_code, expected_exception, expected_message",
        [
            (200, None, None),
            (400, BadRequestError, "BadRequestError"),
            (403, ForbiddenError, "ForbiddenError"),
            (404, NotFoundError, "NotFoundError"),
            (409, ConflictError, "ConflictError"),
            (422, UnprocessableEntityError, "UnprocessableEntityError"),
            (500, InternalServerError, "InternalServerError"),
        ],
    )
    def test_create_workspace(self, httpx_mock: HTTPXMock, status_code, expected_exception, expected_message):
        mock_name = "test-workspace"
        mock_return_value = {
            "id": str(uuid.uuid4()),
            "name": mock_name,
            "inserted_at": datetime.utcnow().isoformat(),
            "updated_at": datetime.utcnow().isoformat(),
        }
        api_url = "http://test_url"
        httpx_mock.add_response(json=mock_return_value, url=f"{api_url}/api/workspaces", status_code=status_code)
        with httpx.Client():
            client = rg.Argilla(api_url=api_url, api_key="admin.apikey")
            if expected_exception:
                with pytest.raises(expected_exception, match=expected_message):
                    ws = rg.Workspace(name="test-workspace", id=uuid.uuid4(), client=client)
                    ws.create()
            else:
                ws = rg.Workspace(name="test-workspace", id=uuid.uuid4(), client=client)
                created_workspace = ws.create()
                assert created_workspace.name == mock_name
                assert created_workspace.id == uuid.UUID(mock_return_value["id"])

    @pytest.mark.parametrize(
        "status_code, expected_exception, expected_message",
        [
            (200, None, None),
            (400, BadRequestError, "BadRequestError"),
            (403, ForbiddenError, "ForbiddenError"),
            (404, NotFoundError, "NotFoundError"),
            (409, ConflictError, "ConflictError"),
            (422, UnprocessableEntityError, "UnprocessableEntityError"),
            (500, InternalServerError, "InternalServerError"),
        ],
    )
    def test_get_workspace(self, httpx_mock: HTTPXMock, status_code, expected_exception, expected_message):
        workspace_id = uuid.uuid4()
        mock_return_value = {
            "id": workspace_id.hex,
            "name": "test-workspace",
            "inserted_at": datetime.utcnow().isoformat(),
            "updated_at": datetime.utcnow().isoformat(),
        }
        api_url = "http://test_url"
        httpx_mock.add_response(
            json=mock_return_value, url=f"{api_url}/api/v1/workspaces/{workspace_id}", status_code=status_code
        )
        with httpx.Client():
            client = rg.Argilla(api_url="http://test_url", api_key="admin.apikey")

            if expected_exception:
                with pytest.raises(expected_exception, match=expected_message):
                    workspace = rg.Workspace(name="test-workspace", id=workspace_id, client=client)
                    workspace = workspace.get()
            else:
                workspace = rg.Workspace(name="test-workspace", id=workspace_id, client=client)
                workspace = workspace.get()
                assert workspace.name == mock_return_value["name"]
                assert workspace.id == workspace_id

    def test_list_workspaces(self, httpx_mock: HTTPXMock):
        mock_return_value = {
            "items": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "test-workspace",
                    "inserted_at": datetime.utcnow().isoformat(),
                    "updated_at": datetime.utcnow().isoformat(),
                },
                {
                    "id": str(uuid.uuid4()),
                    "name": "another-test-workspace",
                    "inserted_at": datetime.utcnow().isoformat(),
                    "updated_at": datetime.utcnow().isoformat(),
                },
            ]
        }
        api_url = "http://test_url"
        httpx_mock.add_response(json=mock_return_value, url=f"{api_url}/api/v1/me/workspaces")
        with httpx.Client():
            client = rg.Argilla(api_url="http://test_url", api_key="admin.apikey")
            workspaces = client.workspaces
        assert len(workspaces) == 2
        for i in range(len(workspaces)):
            assert workspaces[i].name == mock_return_value["items"][i]["name"]
            assert workspaces[i].id == uuid.UUID(mock_return_value["items"][i]["id"])


class TestWorkspacesAPI:
    def test_get_workspace_by_name(self, httpx_mock: HTTPXMock):
        mock_return_value = {
            "items": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "test-workspace",
                    "inserted_at": datetime.utcnow().isoformat(),
                    "updated_at": datetime.utcnow().isoformat(),
                },
                {
                    "id": str(uuid.uuid4()),
                    "name": "other-workspace",
                    "inserted_at": datetime.utcnow().isoformat(),
                    "updated_at": datetime.utcnow().isoformat(),
                },
            ]
        }
        api_url = "http://test_url"
        httpx_mock.add_response(json=mock_return_value, url=f"{api_url}/api/v1/me/workspaces")
        with httpx.Client():
            client = rg.Argilla(api_url=api_url, api_key="admin.apikey")
            ws = client.api.workspaces.get_by_name("test-workspace")
            assert ws is not None
            assert ws.name == "test-workspace"
            assert ws.id == uuid.UUID(mock_return_value["items"][0]["id"])

    def test_multiple_clients_create_workspace(self, httpx_mock: HTTPXMock):
        mock_uuid = str(uuid.uuid4())
        mock_name = "local-test-workspace"
        mock_return = {
            "id": mock_uuid,
            "name": mock_name,
            "inserted_at": datetime.utcnow().isoformat(),
            "updated_at": datetime.utcnow().isoformat(),
        }
        httpx_mock.add_response(
            url="http://localhost:6900/api/workspaces",
            json=mock_return,
        )
        httpx_mock.add_response(
            url="http://argilla.production.net/api/workspaces",
            json=mock_return,
        )
        with httpx.Client():
            local_client = rg.Argilla(api_url="http://localhost:6900", api_key="admin.apikey")
            remote_client = rg.Argilla(api_url="http://argilla.production.net", api_key="admin.apikey")
            assert local_client.api_url == "http://localhost:6900"
            assert remote_client.api_url == "http://argilla.production.net"
            local_workspace = rg.Workspace(name="local-test-workspace", client=local_client)
            local_workspace = local_workspace.create()
            remote_workspace = rg.Workspace(name="remote-test-workspace", client=remote_client)
            remote_workspace = remote_workspace.create()

    def test_delete_workspace(self, httpx_mock: HTTPXMock):
        workspace_id = uuid.uuid4()
        api_url = "http://test_url"
        httpx_mock.add_response(url=f"{api_url}/api/v1/workspaces/{workspace_id}", status_code=204)
        with httpx.Client():
            client = rg.Argilla(api_url=api_url, api_key="admin.apikey")
            client.api.workspaces.delete(workspace_id)

    def test_list_workspace_datasets(self, httpx_mock: HTTPXMock):
        workspace_id = uuid.uuid4()
        mock_return_value = {
            "items": [
                {
                    "id": str(uuid.uuid4()),
                    "name": "test-dataset",
                    "status": "ready",
                    "guidelines": "test-guidelines",
                    "allow_extra_metadata": True,
                    "workspace_id": str(workspace_id),
                    "inserted_at": datetime.utcnow().isoformat(),
                    "updated_at": datetime.utcnow().isoformat(),
                },
                {
                    "id": str(uuid.uuid4()),
                    "name": "another-test-dataset",
                    "status": "ready",
                    "guidelines": "test-guidelines",
                    "allow_extra_metadata": True,
                    "workspace_id": str(workspace_id),
                    "inserted_at": datetime.utcnow().isoformat(),
                    "updated_at": datetime.utcnow().isoformat(),
                },
            ]
        }
        api_url = "http://test_url"
        httpx_mock.add_response(json=mock_return_value, url=f"{api_url}/api/v1/me/datasets")
        with httpx.Client():
            client = rg.Argilla(api_url=api_url, api_key="admin.apikey")
            datasets = client.api.datasets.list(workspace_id)
            assert len(datasets) == 2
            for i in range(len(datasets)):
                assert datasets[i].name == mock_return_value["items"][i]["name"]
                assert datasets[i].id == uuid.UUID(mock_return_value["items"][i]["id"])



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, TYPE_CHECKING, Optional, overload, Union, Sequence
from uuid import UUID

from argilla_sdk._api._workspaces import WorkspacesAPI
from argilla_sdk._helpers import GenericIterator
from argilla_sdk._helpers._mixins import LoggingMixin
from argilla_sdk._models import WorkspaceModel
from argilla_sdk._resource import Resource
from argilla_sdk.client import Argilla

if TYPE_CHECKING:
    from argilla_sdk import Dataset, User


class Workspace(Resource):
    """Class for interacting with Argilla workspaces. Workspaces are used to organize datasets in the Argilla server.

    Attributes:
        name (str): The name of the workspace.
        id (UUID): The ID of the workspace. This is a unique identifier for the workspace in the server.
        datasets (List[Dataset]): A list of all datasets in the workspace.
        users (WorkspaceUsers): A list of all users in the workspace.
    """

    name: Optional[str]

    _api: "WorkspacesAPI"

    def __init__(
        self,
        name: Optional[str] = None,
        id: Optional[UUID] = None,
        client: Optional["Argilla"] = None,
        _model: Optional[WorkspaceModel] = None,
    ) -> None:
        """Initializes a Workspace object with a client and a name or id

        Parameters:
            client (Argilla): The client used to interact with Argilla
            name (str): The name of the workspace
            id (UUID): The id of the workspace
            _model (WorkspaceModel): The internal Pydantic model of the workspace from/to the server
        Returns:
            Workspace: The initialized workspace object
        """
        client = client or Argilla._get_default()
        super().__init__(client=client, api=client.api.workspaces)
        self._sync(model=WorkspaceModel(name=name, id=id) if not _model else _model)

    def exists(self) -> bool:
        """
        Checks if the workspace exists in the server

        Returns:
            bool: True if the workspace exists, False otherwise
        """
        return self._api.exists(self.id)

    def add_user(self, user: Union["User", str]) -> "User":
        """Adds a user to the workspace. After adding a user to the workspace, it will have access to the datasets
        in the workspace.

        Args:
            user (Union[User, str]): The user to add to the workspace. Can be a User object or a username.

        Returns:
            User: The user that was added to the workspace
        """
        return self.users.add(user)

    def remove_user(self, user: Union["User", str]) -> "User":
        """Removes a user from the workspace. After removing a user from the workspace, it will no longer have access

        Args:
            user (Union[User, str]): The user to remove from the workspace. Can be a User object or a username.

        Returns:
            User: The user that was removed from the workspace.
        """
        return self.users.delete(user)

    # TODO: Make this method private
    def list_datasets(self) -> List["Dataset"]:
        from argilla_sdk.datasets import Dataset

        datasets = self._client.api.datasets.list(self.id)
        self.log(f"Got {len(datasets)} datasets for workspace {self.id}")
        return [Dataset.from_model(model=dataset, client=self._client) for dataset in datasets]

    ############################
    # Properties
    ############################

    @property
    def name(self) -> Optional[str]:
        return self._model.name

    @name.setter
    def name(self, value: str) -> None:
        self._model.name = value

    @property
    def datasets(self) -> List["Dataset"]:
        """List all datasets in the workspace

        Returns:
            List[Dataset]: A list of all datasets in the workspace
        """
        return self.list_datasets()

    @property
    def users(self) -> "WorkspaceUsers":
        """List all users in the workspace

        Returns:
            WorkspaceUsers: A list of all users in the workspace

        """
        return WorkspaceUsers(workspace=self)

    ############################
    # Private methods
    ############################


class WorkspaceUsers(Sequence["User"], LoggingMixin):
    class _Iterator(GenericIterator["User"]):
        pass

    def __init__(self, workspace: "Workspace") -> None:
        self._workspace = workspace

    @overload
    def add(self, user: "User") -> "User":
        ...

    @overload
    def add(self, user: str) -> "User":
        ...

    def add(self, user: Union["User", str]) -> "User":
        if isinstance(user, str):
            return self._add_user_by_username(username=user)
        return user.add_to_workspace(workspace=self._workspace)

    @overload
    def delete(self, user: "User") -> "User":
        ...

    @overload
    def delete(self, user: str) -> "User":
        ...

    def delete(self, user: Union["User", str]) -> "User":
        if isinstance(user, str):
            return self._delete_user_by_username(username=user)
        return user.remove_from_workspace(workspace=self._workspace)

    def __iter__(self):
        return self._Iterator(self._list_users())

    def __getitem__(self, index: int) -> "User":
        return self._list_users()[index]

    def __len__(self) -> int:
        return len(self._list_users())

    ####################
    # Private methods
    ####################

    def _list_users(self) -> List["User"]:
        users = self._workspace._client.users.list(workspace=self._workspace)
        self.log(f"Got {len(users)} users for workspace {self._workspace.id}")
        return users

    def _delete_user_by_username(self, username: str) -> "User":
        user = self._workspace._client.users(username=username)
        if not user.exists():
            raise ValueError(f"User {username} does not exist")
        return user.remove_from_workspace(workspace=self._workspace)

    def _add_user_by_username(self, username: str) -> "User":
        user = self._workspace._client.users(username=username)
        if not user.exists():
            raise ValueError(f"User {username} does not exist")
        return user.add_to_workspace(workspace=self._workspace)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from enum import Enum
from typing import Dict, Optional, Union, Any
from uuid import UUID

from pydantic import BaseModel, field_serializer, field_validator, Field


class ResponseStatus(str, Enum):
    draft = "draft"
    submitted = "submitted"
    discarded = "discarded"


class UserResponseModel(BaseModel):
    """Schema for the `FeedbackRecord` user response."""

    values: Union[Dict[str, Dict[str, Any]], None]
    status: ResponseStatus
    user_id: Optional[UUID] = Field(None, validate_default=True)

    class Config:
        validate_assignment = True

    @field_validator("user_id")
    @classmethod
    def user_id_must_have_value(cls, user_id: Optional[UUID]):
        if not user_id:
            warnings.warn(
                "`user_id` not provided, so it will be set to `None`. Which is not an"
                " issue, unless you're planning to log the response in Argilla, as"
                " it will be automatically set to the active `user_id`.",
            )
        return user_id

    @field_serializer("user_id", when_used="always")
    def serialize_user_id(value: UUID) -> str:
        return str(value)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# We skip the flake8 check because we are importing all the models and the import order is important
# flake8: noqa
from argilla_sdk._models._resource import ResourceModel
from argilla_sdk._models._workspace import WorkspaceModel
from argilla_sdk._models._user import UserModel, Role
from argilla_sdk._models._dataset import DatasetModel
from argilla_sdk._models._record._record import RecordModel
from argilla_sdk._models._record._suggestion import SuggestionModel
from argilla_sdk._models._record._response import UserResponseModel, ResponseStatus
from argilla_sdk._models._record._vector import VectorModel
from argilla_sdk._models._record._metadata import MetadataModel, MetadataValue
from argilla_sdk._models._search import (
    SearchQueryModel,
    AndFilterModel,
    FilterModel,
    RangeFilterModel,
    TermsFilterModel,
    ScopeModel,
)
from argilla_sdk._models._settings._fields import (
    TextFieldModel,
    FieldSettings,
    FieldBaseModel,
    VectorFieldModel,
    FieldModel,
)
from argilla_sdk._models._settings._questions import (
    LabelQuestionModel,
    LabelQuestionSettings,
    MultiLabelQuestionModel,
    QuestionBaseModel,
    QuestionModel,
    QuestionSettings,
    RankingQuestionModel,
    RatingQuestionModel,
    SpanQuestionModel,
    SpanQuestionSettings,
    TextQuestionModel,
    TextQuestionSettings,
)
from argilla_sdk._models._settings._metadata import (
    MetadataFieldModel,
    MetadataPropertyType,
    BaseMetadataPropertySettings,
    TermsMetadataPropertySettings,
    NumericMetadataPropertySettings,
    FloatMetadataPropertySettings,
    IntegerMetadataPropertySettings,
)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from datetime import datetime
from uuid import UUID
from typing import Literal

from pydantic import field_serializer

from argilla_sdk._models import ResourceModel

__all__ = ["DatasetModel"]


class DatasetModel(ResourceModel):
    name: str
    status: Literal["draft", "ready"] = "draft"

    guidelines: Optional[str] = None
    allow_extra_metadata: bool = True  # Ideally, the default value should be provided by the server

    workspace_id: Optional[UUID] = None
    last_activity_at: Optional[datetime] = None
    url: Optional[str] = None

    class Config:
        validate_assignment = True
        str_strip_whitespace = True

    @field_serializer("last_activity_at", when_used="unless-none")
    def serialize_last_activity_at(self, value: datetime) -> str:
        return value.isoformat()

    @field_serializer("workspace_id", when_used="unless-none")
    def serialize_workspace_id(self, value: UUID) -> str:
        return str(value)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Generic, List, TypeVar

__all__ = ["GenericIterator"]

Item = TypeVar("Item")


class GenericIterator(Generic[Item]):
    """Generic iterator for any collection of items."""

    def __init__(self, collection: List[Item]):
        self._collection = [v for v in collection]
        self._index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._index < len(self._collection):
            result = self._collection[self._index]
            self._index += 1
            return result
        raise StopIteration



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from datetime import datetime

import httpx
from pytest_httpx import HTTPXMock

import argilla_sdk as rg
from argilla_sdk._models import TextQuestionModel, LabelQuestionModel
from argilla_sdk._models._settings._questions import SpanQuestionModel


class TestQuestionsAPI:
    def test_create_many_questions(self, httpx_mock: HTTPXMock):
        # TODO: Add a test for the delete method in client
        mock_dataset_id = uuid.uuid4()
        mock_return_value = {
            "id": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
            "name": "string",
            "title": "string",
            "required": True,
            "settings": {"type": "text", "use_markdown": False},
            "dataset_id": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
            "inserted_at": datetime.now().isoformat(),
            "updated_at": datetime.now().isoformat(),
        }
        mock_question = {
            "name": "5044cv0wu5",
            "title": "string",
            "description": "string",
            "required": True,
            "settings": {"type": "text", "use_markdown": False},
        }
        mock_question = TextQuestionModel(**mock_question)
        httpx_mock.add_response(
            json=mock_return_value,
            url=f"http://test_url/api/v1/datasets/{mock_dataset_id}/questions",
            method="POST",
            status_code=200,
        )
        with httpx.Client() as client:
            client = rg.Argilla(api_url="http://test_url")
            client.api.questions.create_many(dataset_id=mock_dataset_id, questions=[mock_question])

    def test_create_many_label_questions(self, httpx_mock: HTTPXMock):
        # TODO: Add a test for the delete method in client
        mock_dataset_id = uuid.uuid4()
        mock_return_value = {
            "id": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
            "name": "string",
            "title": "string",
            "required": True,
            "settings": {"type": "labels", "options": [{"text": "positive", "value": "positive"}]},
            "dataset_id": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
            "inserted_at": datetime.now().isoformat(),
            "updated_at": datetime.now().isoformat(),
        }
        mock_question = {
            "name": "5044cv0wu5",
            "title": "string",
            "description": "string",
            "required": True,
            "settings": {
                "type": "label",
                "options": [{"text": "negative", "value": "negative"}, {"text": "positive", "value": "positive"}],
            },
        }
        mock_question = LabelQuestionModel(**mock_question)
        httpx_mock.add_response(
            json=mock_return_value,
            url=f"http://test_url/api/v1/datasets/{mock_dataset_id}/questions",
            method="POST",
            status_code=200,
        )
        with httpx.Client() as client:
            client = rg.Argilla(api_url="http://test_url")
            client.api.questions.create_many(dataset_id=mock_dataset_id, questions=[mock_question])

    def test_create_span_question(self, httpx_mock: HTTPXMock):
        mock_dataset_id = uuid.uuid4()
        mock_return_value = {
            "id": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
            "name": "string",
            "title": "string",
            "required": True,
            "settings": {
                "type": "span",
                "allow_overlapping": True,
                "field": "text",
                "visible_options": 3,
                "options": [
                    {"value": "value 1", "text": "Value 1"},
                    {"value": "value 2", "text": "Value 2"},
                    {"value": "value 3", "text": "Value 3"},
                ],
            },
            "inserted_at": datetime.now().isoformat(),
            "updated_at": datetime.now().isoformat(),
        }

        httpx_mock.add_response(
            json=mock_return_value,
            url=f"http://test_url/api/v1/datasets/{mock_dataset_id}/questions",
            method="POST",
            status_code=200,
        )

        with httpx.Client() as _:
            question = SpanQuestionModel(
                name="5044cv0wu5",
                title="string",
                description="string",
                required=True,
                settings={
                    "type": "span",
                    "allow_overlapping": True,
                    "field": "text",
                    "visible_options": 3,
                    "options": [
                        {"value": "value 1", "text": "Value 1"},
                        {"value": "value 2", "text": "Value 2"},
                        {"value": "value 3", "text": "Value 3"},
                    ],
                },
            )

            client = rg.Argilla(api_url="http://test_url")
            created_question = client.api.questions.create(dataset_id=mock_dataset_id, question=question)
            assert created_question.model_dump(exclude_unset=True) == mock_return_value



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argilla_sdk._models import ResourceModel

import re
from pydantic import field_validator

__all__ = ["WorkspaceModel"]


class WorkspaceModel(ResourceModel):
    name: str

    class Config:
        validate_assignment = True
        str_strip_whitespace = True

    @field_validator("name")
    @classmethod
    def validate_name(cls, value):
        """Validate the name of the workspace is url safe"""
        if not re.match(r"^[a-zA-Z0-9_-]+$", value):
            raise ValueError("Workspace name must be url safe")
        return value



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Optional

from pydantic import field_validator

from argilla_sdk._models import ResourceModel

__all__ = ["UserModel", "Role"]


class Role(str, Enum):
    annotator = "annotator"
    admin = "admin"
    owner = "owner"


class UserModel(ResourceModel):
    username: str
    role: Role = Role.annotator

    first_name: Optional[str] = None
    last_name: Optional[str] = None
    password: Optional[str] = None

    class Config:
        validate_assignment = True
        str_strip_whitespace = True

    @field_validator("first_name")
    @classmethod
    def __validate_first_name(cls, v, values):
        """Set first_name to username if not provided"""
        if isinstance(v, str):
            return v
        elif not v:
            return values["username"]

    @field_validator("username", mode="before")
    @classmethod
    def __validate_username(cls, username: str):
        """Ensure that the username is not empty"""
        if not username:
            raise ValueError("Username cannot be empty")
        return username



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4

import pytest

import argilla_sdk as rg
from argilla_sdk.records._resource import Record


@pytest.fixture
def dataset():
    settings = rg.Settings(
        fields=[rg.TextField(name="prompt")],
        questions=[rg.LabelQuestion(name="label", labels=["negative", "positive"])],
        metadata=[rg.FloatMetadataProperty(name="score")],
        vectors=[rg.VectorField(name="vector", dimensions=3)],
    )
    workspace = rg.Workspace(name="workspace", id=uuid4())
    return rg.Dataset(
        name="test_dataset",
        settings=settings,
        workspace=workspace,
    )


def test_ingest_record_from_dict(dataset):
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "What is the capital of France?",
            "label": "positive",
        },
    )

    assert record.fields.prompt == "What is the capital of France?"
    assert record.suggestions.label.value == "positive"


def test_ingest_record_from_dict_with_mapping(dataset):
    record = Record.from_dict(
        dataset=dataset,
        data={
            "my_prompt": "What is the capital of France?",
            "label": "positive",
        },
        mapping={
            "my_prompt": "prompt",
        },
    )

    assert record.fields.prompt == "What is the capital of France?"
    assert record.suggestions.label.value == "positive"


def test_ingest_record_from_dict_with_suggestions(dataset):
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
        },
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.suggestions.label.value == "negative"


def test_ingest_record_from_dict_with_suggestions_scores(dataset):
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
            "score": 0.9,
            "model": "model_name",
        },
        mapping={
            "score": "label.suggestion.score",
            "model": "label.suggestion.agent",
        },
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.suggestions.label.value == "negative"
    assert record.suggestions.label.score == 0.9
    assert record.suggestions.label.agent == "model_name"


def test_ingest_record_from_dict_with_suggestions_scores_and_agent(dataset):
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
            "score": 0.9,
            "model": "model_name",
        },
        mapping={
            "score": "label.suggestion.score",
            "model": "label.suggestion.agent",
        },
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.suggestions.label.value == "negative"
    assert record.suggestions.label.score == 0.9
    assert record.suggestions.label.agent == "model_name"


def test_ingest_record_from_dict_with_responses(dataset):
    user_id = uuid4()
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
        },
        mapping={
            "label": "label.response",
        },
        user_id=user_id,
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.responses.label[0].value == "negative"
    assert record.responses.label[0].user_id == user_id


def test_ingest_record_from_dict_with_id_as_id(dataset):
    record_id = uuid4()
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
            "id": record_id,
        },
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.id == record_id


def test_ingest_record_from_dict_with_id_and_mapping(dataset):
    record_id = uuid4()
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
            "test_id": record_id,
        },
        mapping={
            "test_id": "id",
        },
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.id == record_id


def test_ingest_record_from_dict_with_metadata(dataset):
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
            "score": 0.9,
        },
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.suggestions.label.value == "negative"
    assert record.metadata["score"] == 0.9


def test_ingest_record_from_dict_with_metadata_and_mapping(dataset):
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
            "test_score": 0.9,
        },
        mapping={
            "test_score": "score",
        },
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.suggestions.label.value == "negative"
    assert record.metadata["score"] == 0.9


def test_ingest_record_from_dict_with_vectors(dataset):
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
            "vector": [1, 2, 3],
        },
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.suggestions.label.value == "negative"
    assert record.vectors.vector == [1, 2, 3]


def test_ingest_record_from_dict_with_vectors_and_mapping(dataset):
    record = Record.from_dict(
        dataset=dataset,
        data={
            "prompt": "Hello World, how are you?",
            "label": "negative",
            "test_vector": [1, 2, 3],
        },
        mapping={
            "test_vector": "vector",
        },
    )

    assert record.fields.prompt == "Hello World, how are you?"
    assert record.suggestions.label.value == "negative"
    assert record.vectors.vector == [1, 2, 3]



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from string import ascii_lowercase

import pytest

from argilla_sdk import Argilla, Dataset, Settings, TextField, TextQuestion, Workspace, LabelQuestion


@pytest.fixture
def workspace(client: Argilla) -> Workspace:
    workspace = client.workspaces("test-workspace")
    if not workspace.exists():
        workspace.create()
    yield workspace

    for dataset in workspace.list_datasets():
        client.api.datasets.delete(dataset.id)

    workspace.delete()


@pytest.fixture
def dataset(client: Argilla, workspace: Workspace) -> Dataset:
    name = "".join(random.choices(ascii_lowercase, k=16))
    settings = Settings(
        fields=[TextField(name="text")],
        questions=[
            TextQuestion(name="comment", use_markdown=False),
            LabelQuestion(name="sentiment", labels=["positive", "negative"], required=False),
        ],
    )
    dataset = Dataset(
        name=name,
        workspace=workspace.name,
        settings=settings,
        client=client,
    )
    dataset.create()
    yield dataset
    dataset.delete()


def test_list_records_with_start_offset(client: Argilla, dataset: Dataset):
    dataset.records.add(
        [
            {"text": "The record text field", "id": 1},
            {"text": "The record text field", "id": 2},
        ]
    )

    records = list(dataset.records(start_offset=1))
    assert len(records) == 1


def test_list_records_with_responses(client: Argilla, dataset: Dataset):
    dataset.records.add(
        [
            {"text": "The record text field", "id": 1, "comment": "The comment", "sentiment": "positive"},
            {"text": "The record text field", "id": 2, "comment": "The comment", "sentiment": "negative"},
        ],
        mapping={
            "comment": "comment.response",
            "sentiment": "sentiment.response",
        },
    )

    records = list(dataset.records(with_responses=True))
    assert len(records) == 2

    assert records[0].responses.comment[0].value == "The comment"
    assert records[0].responses.sentiment[0].value == "positive"

    assert records[1].responses.comment[0].value == "The comment"
    assert records[1].responses.sentiment[0].value == "negative"



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from httpx import HTTPStatusError

from argilla_sdk._exceptions._base import ArgillaErrorBase


class ArgillaAPIError(ArgillaErrorBase):
    pass


class BadRequestError(ArgillaAPIError):
    message = "Bad request to the server"


class ForbiddenError(ArgillaAPIError):
    message = "Forbidden request to the server"


class NotFoundError(ArgillaAPIError):
    message = "Resource or entity not found on the server"


class ConflictError(ArgillaAPIError):
    message = "Conflict with the server. Resource or entity already exists"


class UnprocessableEntityError(ArgillaAPIError):
    message = "Unprocessable entity. The server cannot process the request"


class InternalServerError(ArgillaAPIError):
    message = "Internal server error"


class UnauthorizedError(ArgillaAPIError):
    message = "Unauthorized user request to the server"


def api_error_handler(func):
    """Decorator to handle API errors from ResourceAPI methods
    and raise the appropriate exception.
    Args: func: the request method to decorate

    Example:
    ```python

    @api_error_handler
    def get(self, workspace_id: UUID) -> WorkspaceModel:
        ... # same code as before
    ```
    """

    def _error_switch(status_code: int, error_detail: str):
        switch = {
            400: BadRequestError,
            401: UnauthorizedError,
            403: ForbiddenError,
            404: NotFoundError,
            409: ConflictError,
            422: UnprocessableEntityError,
            500: InternalServerError,
        }
        exception_class = switch.get(status_code, ArgillaAPIError)
        raise exception_class(f"{exception_class.message}. Details: {error_detail}")

    def _handler_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HTTPStatusError as e:
            _error_switch(status_code=e.response.status_code, error_detail=e.response.text)

    return _handler_wrapper



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Literal, Union, List
from uuid import UUID, uuid4

from pydantic import BaseModel, field_serializer


class SuggestionModel(BaseModel):
    """Schema for the suggestions for the questions related to the record."""

    value: Any

    question_name: Optional[str] = None
    type: Optional[Literal["model", "human"]] = None
    score: Union[float, List[float], None] = None
    agent: Optional[str] = None
    id: Optional[UUID] = uuid4()
    question_id: Optional[UUID] = None

    @field_serializer("id", when_used="unless-none")
    def serialize_id(self, value: UUID) -> str:
        return str(value)

    @field_serializer("question_id", when_used="unless-none")
    def serialize_question_id(self, value: UUID) -> str:
        return str(value)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import uuid
from datetime import datetime


import argilla_sdk as rg
from argilla_sdk import Argilla


def test_create_dataset(client):
    workspace = client.workspaces[0]
    mock_dataset_name = f"test_create_dataset{datetime.now().strftime('%Y%m%d%H%M%S')}"
    dataset = rg.Dataset(
        name=mock_dataset_name,
        workspace=workspace,
        settings=rg.Settings(
            fields=[
                rg.TextField(name="text"),
            ],
            questions=[
                rg.TextQuestion(name="response"),
            ],
        ),
        client=client,
    )
    dataset.create()
    gotten_dataset = dataset.get()
    assert dataset.id == gotten_dataset.id
    assert dataset.name == mock_dataset_name


def test_add_records(client):
    mock_dataset_name = f"test_add_records{datetime.now().strftime('%Y%m%d%H%M%S')}"
    mock_data = [
        {
            "text": "Hello World, how are you?",
            "label": "positive",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "label": "positive",
            "id": uuid.uuid4(),
        },
    ]
    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.TextQuestion(name="comment", use_markdown=False),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    dataset.create()
    dataset.records.add(records=mock_data)

    dataset_records = list(dataset.records)

    assert dataset.name == mock_dataset_name
    assert dataset_records[0].id == str(mock_data[0]["id"])
    assert dataset_records[1].id == str(mock_data[1]["id"])
    assert dataset_records[2].id == str(mock_data[2]["id"])
    assert dataset_records[0].fields.text == mock_data[0]["text"]
    assert dataset_records[1].fields.text == mock_data[1]["text"]
    assert dataset_records[2].fields.text == mock_data[2]["text"]


def test_add_dict_records(client: Argilla):
    ws = client.workspaces("new_ws")
    if not ws.exists():
        ws.create()

    ds = client.datasets("new_ds", workspace=ws)
    if ds.exists():
        ds.delete()

    ds.settings = rg.Settings(
        fields=[rg.TextField(name="text")],
        questions=[rg.TextQuestion(name="label")],
    )

    ds.create()

    mock_data = [
        {
            "text": "Hello World, how are you?",
            "label": "positive",
            "id": "1",
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": "2",
        },
        {"text": "Hello World, how are you?", "label": "negative", "id": "3"},
    ]

    # Now the dataset is published and is ready for annotate
    ds.records.add(mock_data)

    for record, data in zip(ds.records, mock_data):
        assert record.id == data["id"]
        assert record.fields.text == data["text"]
        assert "label" not in record.__dict__

    for record, data in zip(ds.records(batch_size=1, with_suggestions=True), mock_data):
        assert record.id == data["id"]
        assert record.suggestions.label.value == data["label"]


def test_add_single_record(client: Argilla):
    new_ws = client.workspaces("new_ws")
    if not new_ws.exists():
        new_ws.create()

    ds = client.datasets("new_ds", workspace=new_ws)
    if ds.exists():
        ds.delete()

    ds.settings = rg.Settings(
        fields=[rg.TextField(name="text")],
        questions=[rg.TextQuestion(name="label")],
    )

    ds.create()

    data = {
        "text": "Hello World, how are you?",
        "label": "positive",
        "id": "1",
    }

    # Now the dataset is published and is ready for annotate
    ds.records.add(data)

    records = list(ds.records)
    assert len(records) == 1

    record = records[0]
    assert record.id == data["id"]
    assert record.fields.text == data["text"]


def test_add_records_with_suggestions(client) -> None:
    mock_dataset_name = f"test_add_record_with_suggestions {datetime.now().strftime('%Y%m%d%H%M%S')}"
    mock_data = [
        {
            "text": "Hello World, how are you?",
            "label": "positive",
            "id": uuid.uuid4(),
            "comment": "I'm doing great, thank you!",
            "topics": ["topic1", "topic2"],
            "topics.score": [0.9, 0.8],
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": uuid.uuid4(),
            "comment": "I'm doing great, thank you!",
            "topics": ["topic3"],
            "topics.score": [0.9],
        },
        {
            "text": "Hello World, how are you?",
            "label": "positive",
            "id": uuid.uuid4(),
            "comment": "I'm doing great, thank you!",
            "topics": ["topic1", "topic2", "topic3"],
            "topics.score": [0.9, 0.8, 0.7],
        },
    ]
    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.TextQuestion(name="comment", use_markdown=False),
            rg.MultiLabelQuestion(name="topics", labels=["topic1", "topic2", "topic3"], labels_order="suggestion"),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    dataset.create()
    dataset.records.add(
        mock_data,
        mapping={
            "comment": "comment.suggestion",
            "topics": "topics.suggestion",
            "topics.score": "topics.suggestion.score",
        },
    )
    assert dataset.name == mock_dataset_name

    dataset_records = list(dataset.records(with_suggestions=True))

    assert dataset_records[0].id == str(mock_data[0]["id"])
    assert dataset_records[0].suggestions.comment.value == "I'm doing great, thank you!"
    assert dataset_records[0].suggestions.comment.score is None
    assert dataset_records[0].suggestions.topics.value == ["topic1", "topic2"]
    assert dataset_records[0].suggestions.topics.score == [0.9, 0.8]

    assert dataset_records[1].fields.text == mock_data[1]["text"]
    assert dataset_records[1].suggestions.comment.value == "I'm doing great, thank you!"
    assert dataset_records[1].suggestions.comment.score is None
    assert dataset_records[1].suggestions.topics.value == ["topic3"]
    assert dataset_records[1].suggestions.topics.score == [0.9]

    assert dataset_records[2].suggestions.comment.value == "I'm doing great, thank you!"
    assert dataset_records[2].suggestions.comment.score is None
    assert dataset_records[2].suggestions.topics.value == ["topic1", "topic2", "topic3"]
    assert dataset_records[2].suggestions.topics.score == [0.9, 0.8, 0.7]


def test_add_records_with_responses(client) -> None:
    mock_dataset_name = f"test_modify_record_responses_locally {uuid.uuid4()}"
    mock_data = [
        {
            "text": "Hello World, how are you?",
            "my_label": "positive",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "my_label": "positive",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "my_label": "negative",
            "id": uuid.uuid4(),
        },
    ]
    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.LabelQuestion(name="label", labels=["positive", "negative"]),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    user = rg.User(
        username=f"test_{random.randint(0, 1000)}",
        first_name="test",
        password="testtesttest",
        client=client,
    )
    user.create()
    dataset.create()
    dataset.records.add(
        records=mock_data,
        user_id=user.id,
        mapping={
            "my_label": "label.response",
        },
    )
    assert dataset.name == mock_dataset_name

    dataset_records = list(dataset.records(with_responses=True))

    for record, mock_record in zip(dataset_records, mock_data):
        assert record.id == str(mock_record["id"])
        assert record.fields.text == mock_record["text"]
        assert record.responses.label[0].value == mock_record["my_label"]
        assert record.responses.label[0].user_id == user.id


def test_add_records_with_responses_and_suggestions(client) -> None:
    mock_dataset_name = f"test_modify_record_responses_locally {uuid.uuid4()}"
    mock_data = [
        {
            "text": "Hello World, how are you?",
            "my_label": "negative",
            "my_guess": "positive",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "my_label": "negative",
            "my_guess": "positive",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "my_label": "negative",
            "my_guess": "positive",
            "id": uuid.uuid4(),
        },
    ]
    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.LabelQuestion(name="label", labels=["positive", "negative"]),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    user = rg.User(
        username=f"test_{random.randint(0, 1000)}",
        first_name="test",
        password="testtesttest",
        client=client,
    )
    user.create()
    dataset.create()
    dataset.records.add(
        records=mock_data,
        user_id=user.id,
        mapping={
            "my_label": "label.response",
            "my_guess": "label.suggestion",
        },
    )
    assert dataset.name == mock_dataset_name

    dataset_records = list(dataset.records(with_suggestions=True))

    assert dataset_records[0].id == str(mock_data[0]["id"])
    assert dataset_records[1].fields.text == mock_data[1]["text"]
    assert dataset_records[2].suggestions.label.value == "positive"
    assert dataset_records[2].responses.label[0].value == "negative"
    assert dataset_records[2].responses.label[0].user_id == user.id


def test_add_records_with_fields_mapped(client) -> None:
    mock_dataset_name = f"test_modify_record_responses_locally {uuid.uuid4()}"
    mock_data = [
        {
            "x": "Hello World, how are you?",
            "my_label": "negative",
            "my_guess": "positive",
            "id": uuid.uuid4(),
            "score": 0.5,
        },
        {
            "x": "Hello World, how are you?",
            "my_label": "negative",
            "my_guess": "positive",
            "id": uuid.uuid4(),
            "score": 0.5,
        },
        {
            "x": "Hello World, how are you?",
            "my_label": "negative",
            "my_guess": "positive",
            "id": uuid.uuid4(),
            "score": 0.5,
        },
    ]
    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.LabelQuestion(name="label", labels=["positive", "negative"]),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    user = rg.User(
        username=f"test_{random.randint(0, 1000)}",
        first_name="test",
        password="testtesttest",
        client=client,
    )
    user.create()
    dataset.create()
    dataset.records.add(
        records=mock_data,
        user_id=user.id,
        mapping={
            "my_label": "label.response",
            "my_guess": "label.suggestion",
            "x": "text",
            "score": "label.suggestion.score",
        },
    )
    assert dataset.name == mock_dataset_name

    dataset_records = list(dataset.records(with_suggestions=True))

    assert dataset_records[0].id == str(mock_data[0]["id"])
    assert dataset_records[1].fields.text == mock_data[1]["x"]
    assert dataset_records[2].suggestions.label.value == "positive"
    assert dataset_records[2].suggestions.label.score == 0.5
    assert dataset_records[2].responses.label[0].value == "negative"
    assert dataset_records[2].responses.label[0].value == "negative"
    assert dataset_records[2].responses.label[0].user_id == user.id


def test_add_records_with_id_mapped(client) -> None:
    mock_dataset_name = f"test_modify_record_responses_locally {uuid.uuid4()}"
    mock_data = [
        {
            "x": "Hello World, how are you?",
            "my_label": "negative",
            "my_guess": "positive",
            "uuid": uuid.uuid4(),
        },
        {
            "x": "Hello World, how are you?",
            "my_label": "negative",
            "my_guess": "positive",
            "uuid": uuid.uuid4(),
        },
        {
            "x": "Hello World, how are you?",
            "my_label": "negative",
            "my_guess": "positive",
            "uuid": uuid.uuid4(),
        },
    ]
    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.LabelQuestion(name="label", labels=["positive", "negative"]),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    user = rg.User(
        username=f"test_{random.randint(0, 1000)}",
        first_name="test",
        password="testtesttest",
        client=client,
    )
    user.create()
    dataset.create()
    dataset.records.add(
        records=mock_data,
        user_id=user.id,
        mapping={"my_label": "label.response", "my_guess": "label.suggestion", "x": "text", "uuid": "id"},
    )
    assert dataset.name == mock_dataset_name

    dataset_records = list(dataset.records(with_suggestions=True))

    assert dataset_records[0].id == str(mock_data[0]["uuid"])
    assert dataset_records[1].fields.text == mock_data[1]["x"]
    assert dataset_records[2].suggestions.label.value == "positive"
    assert dataset_records[2].responses.label[0].value == "negative"
    assert dataset_records[2].responses.label[0].user_id == user.id


def test_add_record_resources(client):
    user_id = client.users[0].id
    mock_dataset_name = f"test_add_records{datetime.now().strftime('%Y%m%d%H%M%S')}"
    mock_resources = [
        rg.Record(
            fields={"text": "Hello World, how are you?"},
            suggestions=[
                rg.Suggestion("label", "positive", score=0.9),
                rg.Suggestion("topics", ["topic1", "topic2"], score=[0.9, 0.8]),
            ],
            responses=[rg.Response("label", "positive", user_id=user_id)],
            id=str(uuid.uuid4()),
        ),
        rg.Record(
            fields={"text": "Hello World, how are you?"},
            suggestions=[
                rg.Suggestion("label", "positive", score=0.9),
                rg.Suggestion("topics", ["topic1", "topic2"], score=[0.9, 0.8]),
            ],
            responses=[rg.Response("label", "positive", user_id=user_id)],
            id=str(uuid.uuid4()),
        ),
        rg.Record(
            fields={"text": "Hello World, how are you?"},
            suggestions=[
                rg.Suggestion("label", "positive", score=0.9),
                rg.Suggestion("topics", ["topic1", "topic2"], score=[0.9, 0.8]),
            ],
            responses=[rg.Response("label", "positive", user_id=user_id)],
            id=str(uuid.uuid4()),
        ),
    ]
    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.LabelQuestion(name="label", labels=["positive", "negative"]),
            rg.MultiLabelQuestion(name="topics", labels=["topic1", "topic2", "topic3"]),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    dataset.create()
    dataset.records.add(records=mock_resources)

    dataset_records = list(dataset.records(with_suggestions=True))

    assert dataset.name == mock_dataset_name

    assert dataset_records[0].id == str(mock_resources[0].id)
    assert dataset_records[0].suggestions.label.value == "positive"
    assert dataset_records[0].suggestions.label.score == 0.9
    assert dataset_records[0].suggestions.topics.value == ["topic1", "topic2"]
    assert dataset_records[0].suggestions.topics.score == [0.9, 0.8]

    assert dataset_records[1].id == str(mock_resources[1].id)
    assert dataset_records[1].suggestions.label.value == "positive"
    assert dataset_records[1].suggestions.label.score == 0.9
    assert dataset_records[1].suggestions.topics.value == ["topic1", "topic2"]
    assert dataset_records[1].suggestions.topics.score == [0.9, 0.8]

    assert dataset_records[2].id == str(mock_resources[2].id)
    assert dataset_records[2].suggestions.label.value == "positive"
    assert dataset_records[2].suggestions.label.score == 0.9
    assert dataset_records[2].suggestions.topics.value == ["topic1", "topic2"]
    assert dataset_records[2].suggestions.topics.score == [0.9, 0.8]


def test_add_records_with_responses_and_same_schema_name(client: Argilla):
    mock_dataset_name = f"test_modify_record_responses_locally {uuid.uuid4()}"
    mock_data = [
        {
            "text": "Hello World, how are you?",
            "label": "negative",
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
        },
    ]
    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.LabelQuestion(name="label", labels=["positive", "negative"]),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    user = rg.User(
        username=f"test_{random.randint(0, 1000)}",
        first_name="test",
        password="testtesttest",
        client=client,
    )
    user.create()
    dataset.create()
    dataset.records.add(
        records=mock_data,
        user_id=user.id,
        mapping={"label": "label.response", "text": "text"},
    )
    assert dataset.name == mock_dataset_name

    dataset_records = list(dataset.records(with_responses=True))

    assert dataset_records[0].fields.text == mock_data[1]["text"]
    assert dataset_records[1].responses.label[0].value == "negative"
    assert dataset_records[1].responses.label[0].user_id == user.id


def test_add_records_objects_with_responses(client: Argilla):
    mock_dataset_name = f"test_modify_record_responses_locally {uuid.uuid4()}"

    settings = rg.Settings(
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.TextQuestion(name="comment", use_markdown=False),
            rg.LabelQuestion(name="label", labels=["positive", "negative"]),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        settings=settings,
        client=client,
    )
    user = rg.User(
        username=f"test_{random.randint(0, 1000)}",
        first_name="test",
        password="testtesttest",
        client=client,
    )
    user.create()
    dataset.create()

    records = [
        rg.Record(
            fields={"text": "Hello World, how are you?"},
            responses=[rg.Response("label", "negative", user_id=user.id)],
            id=str(uuid.uuid4()),
        ),
        rg.Record(
            fields={"text": "Hello World, how are you?"},
            responses=[rg.Response("label", "positive", user_id=user.id)],
            id=str(uuid.uuid4()),
        ),
        rg.Record(
            fields={"text": "Hello World, how are you?"},
            responses=[rg.Response("comment", "The comment", user_id=user.id)],
            id=str(uuid.uuid4()),
        ),
    ]

    dataset.records.add(records)

    dataset_records = list(dataset.records())

    assert dataset.name == mock_dataset_name
    assert dataset_records[0].id == records[0].id
    assert dataset_records[0].responses.label[0].value == "negative"

    assert dataset_records[1].id == records[1].id
    assert dataset_records[1].responses.label[0].value == "positive"

    assert dataset_records[2].id == records[2].id
    assert dataset_records[2].responses.comment[0].value == "The comment"



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from collections import defaultdict
from typing import Any, Dict, List, TYPE_CHECKING, Union, Iterable

if TYPE_CHECKING:
    from argilla_sdk import Record


class GenericExportMixin(Iterable["Record"], ABC):
    """This is a mixin class for DatasetRecords and Export classes.
    It handles methods for exporting records to generic python formats."""

    def to_dict(self, flatten: bool = False, orient: str = "names") -> Dict[str, Any]:
        """Export records to a dictionary with either names or record index as keys.
        Args:
            flatten (bool): The structure of the exported dictionary.
                - True: The record fields, metadata, suggestions and responses will be flattened.
                - False: The record fields, metadata, suggestions and responses will be nested.
            orient (str): The orientation of the exported dictionary.
                - "names": The keys of the dictionary will be the names of the fields, metadata, suggestions and responses.
                - "index": The keys of the dictionary will be the id of the records.
        Returns:
            dataset_records (Dict[str, Union[str, float, int, list]]): The exported records in a dictionary format.
        """
        records = [r for r in self]
        return self._export_to_dict(records=records, flatten=flatten, orient=orient)

    def to_list(self, flatten: bool = False) -> List[Dict[str, Any]]:
        """Export records to a list of dictionaries with either names or record index as keys.
        Args:
            flatten (bool): The structure of the exported dictionary.
                - True: The record fields, metadata, suggestions and responses will be flattened.
                - False: The record fields, metadata, suggestions and responses will be nested.
        Returns:
            dataset_records (List[Dict[str, Union[str, float, int, list]]]): The exported records in a list of dictionaries format.
        """
        records = [r for r in self]
        return self._export_to_list(records=records, flatten=flatten)

    ############################
    # Private methods
    ############################

    def _export_to_dict(
        self, records: List["Record"], flatten=False, orient="names"
    ) -> Dict[str, Union[str, float, int, list]]:
        if orient == "names":
            dataset_records: dict = defaultdict(list)
            for record in records:
                for key, value in self._record_to_dict(record=record, flatten=flatten).items():
                    dataset_records[key].append(value)
        elif orient == "index":
            dataset_records: dict = {}
            for record in records:
                dataset_records[record.id] = self._record_to_dict(record=record, flatten=flatten)
        else:
            raise ValueError(f"Invalid value for orient parameter: {orient}")
        return dict(dataset_records)

    def _export_to_list(self, records: List["Record"], flatten=False) -> List[Dict[str, Union[str, float, int, list]]]:
        dataset_records: list = []
        for record in records:
            dataset_records.append(self._record_to_dict(record=record, flatten=flatten))
        return dataset_records

    @staticmethod
    def _record_to_dict(record: "Record", flatten=False) -> Dict[str, Any]:
        """Converts a Record object to a dictionary for export.
        Args:
            record (Record): The Record object to convert.
            flatten (bool): The structure of the exported dictionary.
                - True: The record fields, metadata, suggestions and responses will be flattened
                        so that their keys becomes the keys of the record dictionary, using
                        dot notation for nested keys. i.e. `label.suggestion` and `label.response`
                - False: The record fields, metadata, suggestions and responses will be nested as
                        dictionaries within the record dictionary. i.e. `label: {suggestion: ..., response: ...}`
        Returns:
            A dictionary representing the record.
        """
        record_dict = record.to_dict()
        if flatten:
            responses: dict = record_dict.pop("responses")
            suggestions: dict = record_dict.pop("suggestions")
            fields: dict = record_dict.pop("fields")
            metadata: dict = record_dict.pop("metadata")
            record_dict.update(fields)
            record_dict.update(metadata)
            question_names = set(suggestions.keys()).union(responses.keys())
            for question_name in question_names:
                _suggestion: Union[Dict, None] = suggestions.get(question_name)
                if _suggestion:
                    record_dict[f"{question_name}.suggestion"] = _suggestion.pop("value")
                    record_dict.update(
                        {f"{question_name}.suggestion.{key}": value for key, value in _suggestion.items()}
                    )
                for _response in responses.get(question_name, []):
                    user_id = _response.pop("user_id")
                    record_dict[f"{question_name}.response.{user_id}"] = _response.pop("value")
                    record_dict.update(
                        {f"{question_name}.response.{user_id}.{key}": value for key, value in _response.items()}
                    )
        return record_dict



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from argilla_sdk import (
    Argilla,
    Settings,
    TextField,
    TextQuestion,
    SpanQuestion,
    LabelQuestion,
    MultiLabelQuestion,
    RatingQuestion,
    RankingQuestion,
    TermsMetadataProperty,
    IntegerMetadataProperty,
    FloatMetadataProperty,
)


def test_publish_dataset(client: "Argilla"):
    new_ws = client.workspaces("new_ws")
    if not new_ws.exists():
        new_ws.create()

    assert new_ws.exists(), "The workspace was not created"

    ds = client.datasets("new_ds", workspace=new_ws)
    if ds.exists():
        ds.delete()

    assert not ds.exists(), "The dataset was not deleted"

    ds.settings = Settings(
        guidelines="This is a test dataset",
        allow_extra_metadata=True,
        fields=[TextField(name="text-field")],
        questions=[
            TextQuestion(name="text-question"),
            RatingQuestion(name="rating-question", values=[1, 2, 3, 4, 5]),
            RankingQuestion(name="ranking-question", values=["rank1", "rank2", "rank3"]),
            LabelQuestion(name="label-question", labels=["A", "B", "C"]),
            MultiLabelQuestion(name="multi-label-question", labels=["A", "B", "C"]),
            SpanQuestion(name="span-question", field="text-field", labels=["label1", "label2"]),
        ],
        metadata=[
            TermsMetadataProperty(name="metadata-property", options=["term1", "term2"]),
            TermsMetadataProperty(name="term-property"),
            IntegerMetadataProperty(name="metadata-property-2", min=0, max=10),
            FloatMetadataProperty(name="metadata-property-3", min=0, max=10),
        ],
    )

    ds.create()

    created_dataset = client.datasets(name=ds.name, workspace=new_ws)
    assert created_dataset.exists(), "The dataset was not found"
    assert created_dataset == ds
    assert created_dataset.settings == ds.settings, "The settings were not saved"

    assert created_dataset.guidelines == ds.guidelines
    assert created_dataset.allow_extra_metadata == ds.allow_extra_metadata
    assert created_dataset.fields == ds.fields
    assert created_dataset.questions == ds.questions
    assert created_dataset.schema == ds.schema



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

from argilla_sdk._models import FieldModel, FieldSettings, MetadataFieldModel, TextFieldModel, VectorFieldModel
from argilla_sdk.settings._common import SettingsPropertyBase
from argilla_sdk.settings._metadata import MetadataField

__all__ = ["TextField", "FieldType", "VectorField"]


class TextField(SettingsPropertyBase):
    """Text field for use in Argilla `Dataset` `Settings`"""

    _model: TextFieldModel

    def __init__(
        self,
        name: str,
        title: Optional[str] = None,
        use_markdown: Optional[bool] = False,
        required: Optional[bool] = True,
        description: Optional[str] = None,
    ) -> None:
        """Text field for use in Argilla `Dataset` `Settings`
        Parameters:
            name (str): The name of the field
            title (Optional[str], optional): The title of the field. Defaults to None.
            use_markdown (Optional[bool], optional): Whether to use markdown. Defaults to False.
            required (Optional[bool], optional): Whether the field is required. Defaults to True.
            description (Optional[str], optional): The description of the field. Defaults to None.

        """
        self._model = TextFieldModel(
            name=name,
            title=title,
            required=required or True,
            description=description,
            settings=FieldSettings(type="text", use_markdown=use_markdown),
        )

    @property
    def use_markdown(self) -> Optional[bool]:
        return self._model.settings.use_markdown

    @classmethod
    def from_model(cls, model: TextFieldModel) -> "TextField":
        instance = cls(name=model.name)
        instance._model = model

        return instance

    @classmethod
    def from_dict(cls, data: dict) -> "TextField":
        model = TextFieldModel(**data)
        return cls.from_model(model=model)


class VectorField(SettingsPropertyBase):
    """Vector field for use in Argilla `Dataset` `Settings`"""

    _model: VectorFieldModel

    def __init__(
        self,
        name: str,
        dimensions: int,
        title: Optional[str] = None,
    ) -> None:
        """Vector field for use in Argilla `Dataset` `Settings`

        Parameters:
            name (str): The name of the field
            dimensions (int): The number of dimensions in the vector
            title (Optional[str], optional): The title of the field. Defaults to None.
        """
        self._model = VectorFieldModel(
            name=name,
            title=title,
            dimensions=dimensions,
        )

    @classmethod
    def from_model(cls, model: VectorFieldModel) -> "VectorField":
        instance = cls(name=model.name, dimensions=model.dimensions)
        instance._model = model

        return instance

    @classmethod
    def from_dict(cls, data: dict) -> "VectorField":
        model = VectorFieldModel(**data)
        return cls.from_model(model=model)

    @property
    def dimensions(self) -> int:
        return self._model.dimensions

    @property
    def title(self) -> Optional[str]:
        return self._model.title

    @property
    def name(self) -> str:
        return self._model.name


FieldType = Union[TextField, VectorField, MetadataField]


def field_from_model(model: FieldModel) -> FieldType:
    """Create a field instance from a field model"""
    if isinstance(model, TextFieldModel):
        return TextField.from_model(model)
    elif isinstance(model, VectorFieldModel):
        return VectorField.from_model(model)
    elif isinstance(model, MetadataFieldModel):
        return MetadataField.from_model(model)
    else:
        raise ValueError(f"Unsupported field model type: {type(model)}")


def field_from_dict(data: dict) -> FieldType:
    """Create a field instance from a field dictionary"""
    if data["type"] == "text":
        return TextField.from_dict(data)
    elif data["type"] == "vector":
        return VectorField.from_dict(data)
    elif data["type"] == "metadata":
        return MetadataField.from_dict(data)
    else:
        raise ValueError(f"Unsupported field type: {data['type']}")
